# Copyright 2021-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT


import pandas as pd
import os

import caliperreader as cr

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame
from hatchet.util.timer import Timer


class CaliperNativeReader:
    """Read in a native `.cali` file using Caliper's python reader."""

    def __init__(self, filename_or_caliperreader):
        """Read in a native cali using Caliper's python reader.

        Args:
            filename_or_caliperreader (str or CaliperReader): name of a `cali` file OR
                a CaliperReader object
        """
        self.filename_or_caliperreader = filename_or_caliperreader
        self.filename_ext = ""

        self.df_nodes = {}
        self.metric_cols = []
        self.record_data_cols = []
        self.node_dicts = []
        self.callpath_to_node = {}
        self.idx_to_node = {}
        self.callpath_to_idx = {}
        self.global_nid = 0

        self.default_metric = None

        self.timer = Timer()

        if isinstance(self.filename_or_caliperreader, str):
            _, self.filename_ext = os.path.splitext(filename_or_caliperreader)

    def read_metrics(self, ctx="path"):
        all_metrics = []
        records = self.filename_or_caliperreader.records

        # read metadata from the caliper reader
        for record in records:
            node_dict = {}
            if ctx in record:
                # get the node label and callpath for the record
                if isinstance(record[ctx], list):
                    # specify how to parse cupti records
                    if "cupti.activity.kind" in record:
                        if record["cupti.activity.kind"] == "kernel":
                            node_label = record["cupti.kernel.name"]
                            node_callpath = tuple(record[ctx] + [node_label])
                        elif record["cupti.activity.kind"] == "memcpy":
                            node_label = record["cupti.activity.kind"]
                            node_callpath = tuple(record[ctx] + [node_label])
                    else:
                        node_label = record[ctx][-1]
                        node_callpath = tuple(record[ctx])
                else:
                    node_label = record[ctx][-1]
                    node_callpath = tuple([record[ctx]])

                # get node nid based on callpath
                node_dict["nid"] = self.callpath_to_idx.get(node_callpath)

                for item in record.keys():
                    if item != ctx:
                        if item not in self.record_data_cols:
                            self.record_data_cols.append(item)

                        if (
                            self.filename_or_caliperreader.attribute(
                                item
                            ).attribute_type()
                            == "double"
                        ):
                            node_dict[item] = float(record[item])
                        elif (
                            self.filename_or_caliperreader.attribute(
                                item
                            ).attribute_type()
                            == "int"
                        ):
                            node_dict[item] = int(record[item])
                        elif item == "function":
                            if isinstance(record[item], list):
                                node_dict[item] = record[item][-1]
                            else:
                                node_dict[item] = record[item]
                        else:
                            node_dict[item] = record[item]

                all_metrics.append(node_dict)

        # make list of metric columns
        for col in self.record_data_cols:
            if self.filename_or_caliperreader.attribute(col).is_value():
                self.metric_cols.append(col)

        df_metrics = pd.DataFrame.from_dict(data=all_metrics)
        return df_metrics

    def create_graph(self, ctx="path"):
        def _create_parent(child_node, parent_callpath):
            """We may encounter a parent node in the callpath before we see it
            as a child node. In this case, we need to create a hatchet node for
            the parent.

            This function recursively creates parent nodes in a callpath
            until it reaches the already existing parent in that callpath.
            """
            parent_node = self.callpath_to_node.get(parent_callpath)

            if parent_node:
                # return if arrives at the parent
                parent_node.add_child(child_node)
                child_node.add_parent(parent_node)
                return
            else:
                # else create the parent and add parent/child
                grandparent_callpath = parent_callpath[:-1]
                parent_name = parent_callpath[-1]

                parent_node = Node(
                    Frame({"type": "function", "name": parent_name}), None
                )

                self.callpath_to_node[parent_callpath] = parent_node
                self.callpath_to_idx[parent_callpath] = self.global_nid

                node_dict = dict(
                    {"name": parent_name, "node": parent_node, "nid": self.global_nid},
                )
                self.idx_to_node[self.global_nid] = node_dict
                self.global_nid += 1

                parent_node.add_child(child_node)
                child_node.add_parent(parent_node)
                _create_parent(parent_node, grandparent_callpath)

        list_roots = []
        parent_hnode = None
        records = self.filename_or_caliperreader.records

        for record in records:
            node_label = ""
            if ctx in record:
                # if it's a list, then it's a callpath
                if isinstance(record[ctx], list):
                    # specify how to parse cupti records
                    if "cupti.activity.kind" in record:
                        if record["cupti.activity.kind"] == "kernel":
                            node_label = record["cupti.kernel.name"]
                            node_callpath = tuple(record[ctx] + [node_label])
                            parent_callpath = node_callpath[:-1]
                            node_type = "kernel"
                        elif record["cupti.activity.kind"] == "memcpy":
                            node_label = record["cupti.activity.kind"]
                            node_callpath = tuple(record[ctx] + [node_label])
                            parent_callpath = node_callpath[:-1]
                            node_type = "memcpy"
                        else:
                            Exception("Haven't seen this activity kind yet")
                    else:
                        node_label = record[ctx][-1]
                        node_callpath = tuple(record[ctx])
                        parent_callpath = node_callpath[:-1]
                        node_type = "function"

                    hnode = self.callpath_to_node.get(node_callpath)

                    if not hnode:
                        frame = Frame({"type": node_type, "name": node_label})
                        hnode = Node(frame, None)
                        self.callpath_to_node[node_callpath] = hnode

                        # get parent from node callpath
                        parent_hnode = self.callpath_to_node.get(parent_callpath)

                        # create parent if it doesn't exist
                        # else if parent already exists, add child-parent
                        if not parent_hnode:
                            _create_parent(hnode, parent_callpath)
                        else:
                            parent_hnode.add_child(hnode)
                            hnode.add_parent(parent_hnode)

                        self.callpath_to_idx[node_callpath] = self.global_nid
                        node_dict = dict(
                            {"name": node_label, "node": hnode, "nid": self.global_nid},
                        )
                        self.idx_to_node[self.global_nid] = node_dict
                        self.global_nid += 1

                # if it's a string, then it's a root
                else:
                    root_label = record[ctx]
                    root_callpath = tuple([root_label])

                    if root_callpath not in self.callpath_to_node:
                        # create the root since it doesn't exist
                        frame = Frame({"type": "function", "name": root_label})
                        graph_root = Node(frame, None)

                        # store callpaths to identify the root
                        self.callpath_to_node[root_callpath] = graph_root
                        self.callpath_to_idx[root_callpath] = self.global_nid
                        list_roots.append(graph_root)

                        node_dict = dict(
                            {
                                "name": root_label,
                                "node": graph_root,
                                "nid": self.global_nid,
                            }
                        )

                        self.idx_to_node[self.global_nid] = node_dict
                        self.global_nid += 1

        return list_roots

    def read(self):
        """Read the caliper records to extract the calling context tree."""
        if isinstance(self.filename_or_caliperreader, str):
            if self.filename_ext != ".cali":
                raise ValueError("from_caliperreader() needs a .cali file")
            else:
                cali_file = self.filename_or_caliperreader
                self.filename_or_caliperreader = cr.CaliperReader()
                self.filename_or_caliperreader.read(cali_file)

        with self.timer.phase("graph construction"):
            list_roots = self.create_graph()
        self.df_nodes = pd.DataFrame(data=list(self.idx_to_node.values()))

        # create a graph object once all the nodes have been added
        graph = Graph(list_roots)
        graph.enumerate_traverse()

        with self.timer.phase("read metrics"):
            df_fixed_data = self.read_metrics()

        metrics = pd.DataFrame.from_dict(data=df_fixed_data)

        # add missing intermediate nodes to the df_fixed_data dataframe
        if "mpi.rank" in df_fixed_data.columns:
            num_ranks = metrics["mpi.rank"].max() + 1
            rank_list = range(0, num_ranks)

        # create a standard dict to be used for filling all missing rows
        default_metric_dict = {}
        for idx, col in enumerate(self.record_data_cols):
            if self.filename_or_caliperreader.attribute(col).is_value():
                default_metric_dict[list(self.record_data_cols)[idx]] = 0
            else:
                default_metric_dict[list(self.record_data_cols)[idx]] = None

        # create a list of dicts, one dict for each missing row
        missing_nodes = []
        for iteridx, row in self.df_nodes.iterrows():
            # check if df_nodes row exists in df_fixed_data
            metric_rows = df_fixed_data.loc[metrics["nid"] == row["nid"]]
            if "mpi.rank" not in self.metric_cols:
                if metric_rows.empty:
                    # add a single row
                    node_dict = dict(default_metric_dict)
                    missing_nodes.append(node_dict)
            else:
                if metric_rows.empty:
                    # add a row per MPI rank
                    for rank in rank_list:
                        node_dict = dict(default_metric_dict)
                        node_dict["nid"] = row["nid"]
                        node_dict["mpi.rank"] = rank
                        missing_nodes.append(node_dict)
                elif len(metric_rows) < num_ranks:
                    # add a row for each missing MPI rank
                    present_ranks = metric_rows["mpi.rank"].values
                    missing_ranks = [x for x in rank_list if x not in present_ranks]
                    for rank in missing_ranks:
                        node_dict = dict(default_metric_dict)
                        node_dict["nid"] = row["nid"]
                        node_dict["mpi.rank"] = rank
                        missing_nodes.append(node_dict)

        df_missing = pd.DataFrame.from_dict(data=missing_nodes)
        df_metrics = pd.concat([df_fixed_data, df_missing], sort=False)

        # dict mapping old to new column names to make columns consistent with
        # other readers
        old_to_new = {
            "mpi.rank": "rank",
            "module#cali.sampler.pc": "module",
            "sum#time.duration": "time",
            "sum#avg#sum#time.duration": "time",
            "inclusive#sum#time.duration": "time (inc)",
            "sum#avg#inclusive#sum#time.duration": "time (inc)",
        }

        # change column names
        new_cols = []
        for col in df_metrics.columns:
            if col in old_to_new:
                new_cols.append(old_to_new[col])
            else:
                new_cols.append(col)
        df_metrics.columns = new_cols

        # create list of exclusive and inclusive metric columns
        exc_metrics = []
        inc_metrics = []
        for column in self.metric_cols:
            # ignore rank as an exc or inc metric
            if column == "mpi.rank":
                continue

            # add new column names to list of metrics if inc or inclusive in
            # old column names
            if "(inc)" in column or "inclusive" in column:
                if column in old_to_new:
                    column = old_to_new[column]
                inc_metrics.append(column)
            else:
                if column in old_to_new:
                    column = old_to_new[column]
                exc_metrics.append(column)

        with self.timer.phase("data frame"):
            # merge the metrics and node dataframes on the nid column
            dataframe = pd.merge(df_metrics, self.df_nodes, on="nid")

            # set the index to be a MultiIndex
            indices = ["node"]
            if "rank" in dataframe.columns:
                indices.append("rank")
            dataframe.set_index(indices, inplace=True)
            dataframe.sort_index(inplace=True)

        # set the default metric
        if self.default_metric is None:
            if "time (inc)" in dataframe.columns:
                self.default_metric = "time"
            elif "avg#inclusive#sum#time.duration" in dataframe.columns:
                self.default_metric = "avg#inclusive#sum#time.duration"
            elif len(inc_metrics) > 0:
                self.default_metric = inc_metrics[0]
            elif len(exc_metrics) > 0:
                self.default_metric = exc_metrics[0]

        metadata = self.filename_or_caliperreader.globals

        return hatchet.graphframe.GraphFrame(
            graph,
            dataframe,
            exc_metrics,
            inc_metrics,
            self.default_metric,
            metadata=metadata,
        )



# Copyright 2022-2024 The University of Arizona and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import inspect
import os
import getpass
import json
from datetime import datetime
import hatchet


def isJsonable(var):
    try:
        json.dumps(var)
        return True
    except TypeError:
        return False


class Log(object):
    def __init__(self, filename="hatchet.log", active=False):
        self._log_file = filename
        self._active = active

        # ensures we only log api calls made explicitly by
        # a user
        self._nested = False

    def set_output_file(self, filename=""):
        self._log_file = filename

    def set_active(self):
        self._active = True

    def set_inactive(self):
        self._active = False

    def append_to_file(self, log):
        """Manages the opening and writing of log information to a file."""

        with open(self._log_file, "a") as f:
            try:
                f.write(json.dumps(log) + "\n")
            except TypeError as e:
                raise e

    def loggable(self, function):
        """A decrator which logs calls to hatchet functions"""

        def inner(*args, **kwargs):
            try:
                # turn on logger
                if (
                    "logging" in kwargs
                    and kwargs["logging"] is True
                    and not self._active
                ):
                    self._active = True

                    self._log_file = "hatchet_{}.log".format(
                        datetime.now().replace(microsecond=0).isoformat()
                    )
                    print(
                        """Notification: Local logging of Hatchet function calls has been enabled. Output will be dumped to {}""".format(
                            self._log_file
                        )
                    )

                elif "logging" in kwargs and kwargs["logging"] is False:
                    self._active = False
                    print(
                        """Notification: Local logging of Hatchet function calls has been disabled."""
                    )

                # if logger is on and user called function
                if self._active and not self._nested:
                    log_dict = {}
                    arg_list = []
                    serlizable_kwargs = {}

                    # Get a user id
                    try:
                        log_dict["user_id"] = os.getuid()
                    except AttributeError:
                        # for windows machines
                        log_dict["user_id"] = getpass.getuser()

                    for i, arg in enumerate(args):
                        if inspect.isfunction(arg):
                            # log functions passed (eg. lambda passed to filter) as source code
                            arg_source = inspect.getsource(arg)
                            arg_list.append(arg_source)

                        elif isinstance(arg, hatchet.GraphFrame):
                            # log a graphframe as a dictionary of metadata
                            graphframe_metadata = {}

                            graphframe_metadata["object"] = arg.__class__.__name__
                            graphframe_metadata["id"] = id(arg)
                            graphframe_metadata["rows"] = arg.dataframe.shape[0]
                            graphframe_metadata["nodes"] = len(arg.graph)

                            arg_list.append(graphframe_metadata)
                        else:
                            # log everything else
                            if isJsonable(arg):
                                arg_list.append(arg)
                            else:
                                arg_list.append(arg.__repr__())

                    log_dict["start"] = datetime.now().isoformat()

                    self._nested = True
                    holder = function(*args, **kwargs)
                    self._nested = False

                    log_dict["end"] = datetime.now().isoformat()
                    log_dict["function"] = function.__name__
                    log_dict["args"] = tuple(arg_list)

                    for key in kwargs:
                        if isJsonable(kwargs[key]):
                            serlizable_kwargs[key] = kwargs[key]
                        else:
                            serlizable_kwargs[key] = kwargs[key].__repr__()

                    log_dict["kwargs"] = serlizable_kwargs

                    self.append_to_file(log_dict)

                    return holder
                else:
                    return function(*args, **kwargs)

            # If there is a file io error when logging
            # we run function as normal and abandon log
            except IOError:
                return function(*args, **kwargs)

        return inner

    # TODO: add log reader


Logger = Log()



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pytest

from hatchet.frame import Frame


def test_constructors():
    f1 = Frame({"name": "foo", "file": "bar.c"})
    f2 = Frame(name="foo", file="bar.c")
    f3 = Frame({"name": "foo"}, file="bar.c")
    f4 = Frame({"name": "foo", "file": "baz.h"}, file="bar.c")

    assert f1 == f2 == f3 == f4


def test_no_attrs():
    with pytest.raises(ValueError):
        Frame({})

    with pytest.raises(ValueError):
        Frame()


def test_comparison():
    assert Frame(a=1) == Frame(a=1)

    assert Frame(a=1) < Frame(a=2)
    assert Frame(a=1) != Frame(a=2)
    assert Frame(a=2) > Frame(a=1)

    assert Frame(a=1) < Frame(b=1)
    assert Frame(a=1) != Frame(b=1)
    assert Frame(b=1) > Frame(a=1)

    assert Frame(a=1, b=1) < Frame(b=1, c=1)
    assert Frame(a=1, b=1) != Frame(b=1, c=1)
    assert Frame(b=1, c=1) > Frame(a=1, b=1)


def test_copy():
    f = Frame(a=1)
    assert f == f.copy()
    assert not (f != f.copy())
    assert not (f > f.copy())
    assert not (f < f.copy())

    f = Frame(a="foo", b="bar", c="baz")
    assert f == f.copy()
    assert not (f != f.copy())
    assert not (f > f.copy())
    assert not (f < f.copy())


def test_getitem():
    f = Frame(a=1, b=2, c=3)
    assert f["a"] == 1
    assert f["b"] == 2
    assert f["c"] == 3

    assert f.get("a") == 1
    assert f.get("b") == 2
    assert f.get("c") == 3

    assert f.get("d") is None
    assert f.get("d", "foo") == "foo"

    with pytest.raises(KeyError):
        assert f["d"]


def test_values():
    f = Frame(a=1, b=2, c=3)
    assert f.values("a") == 1
    assert f.values(["b"]) == (2,)
    assert f.values(("a", "a", "a")) == (1, 1, 1)
    assert f.values(["a", "b"]) == (1, 2)
    assert f.values(["a", "b", "c"]) == (1, 2, 3)
    assert f.values(["b", "c"]) == (2, 3)
    assert f.values("c") == 3

    assert f.values(["foo", "a", "bar", "c"]) == (None, 1, None, 3)


def test_repr():
    assert (
        repr(Frame(foo="baz", bar="quux"))
        == "Frame({'bar': 'quux', 'foo': 'baz', 'type': 'None'})"
    )


def test_str():
    assert (
        str(Frame(foo="baz", bar="quux"))
        == "{'bar': 'quux', 'foo': 'baz', 'type': 'None'}"
    )



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np

from hatchet import GraphFrame
from hatchet.external.console import ConsoleRenderer

import pytest

# check if the pycubexr package is available
# without importing the package.
# there are other ways to do it for python 3+
import importlib

pycubexr_avail = True
try:
    importlib.util.find_spec("pycubexr")
except ImportError:
    pycubexr_avail = False


procedures = [
    "cpi",
    "main",
    "MPI_Init",
    "MPI_Comm_size",
    "MPI_Comm_rank",
    "iteration",
    "MPI_Bcast",
    "MPI_Reduce",
    "iteration.cold.1",
    "MPI_Finalize",
]


@pytest.mark.skipif(not pycubexr_avail, reason="pycubexr package not available")
def test_graphframe(scorep_profile_cubex):
    gf = GraphFrame.from_scorep(str(scorep_profile_cubex))

    assert len(gf.dataframe.groupby("name")) == 10

    for col in gf.dataframe.columns:
        if col in (
            "max time (inc)",
            "min time (inc)",
            "time",
        ):
            assert gf.dataframe[col].dtype == np.float64
        elif col in (
            "hits",
            "end_line",
            "line",
            "visits",
            "bytes_sent",
            "bytes_received",
        ):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "file", "node"):
            assert gf.dataframe[col].dtype == object


@pytest.mark.skipif(not pycubexr_avail, reason="pycubexr package not available")
def test_tree(scorep_profile_cubex):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_scorep(str(scorep_profile_cubex))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "5.056 cpi" in output
    assert "0.507 MPI_Init MPI" in output
    assert "4.539 iteration /p/lustre1/cankur1/test/scorep/cpi.c" in output

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="visits",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "1.000 cpi" in output
    assert "100082.000 iteration /p/lustre1/cankur1/test/scorep/cpi.c" in output


@pytest.mark.skipif(not pycubexr_avail, reason="pycubexr package not available")
def test_graphframe_to_literal(scorep_profile_cubex):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_scorep(str(scorep_profile_cubex))
    graph_literal = gf.to_literal()

    gf2 = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf2.graph)



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import numpy as np
from hatchet.graphframe import GraphFrame
from hatchet.chopper import Chopper


def test_flat_profile(calc_pi_hpct_db):
    """Validate that the flat profile works correctly."""
    graphframe = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    original_rows = graphframe.dataframe["name"].unique()
    flat_profile = graphframe.flat_profile()

    # Check if the node names are exactly the same
    assert sorted(flat_profile.index.tolist()) == sorted(original_rows)

    # Check if the aggregated time is correct for the root node.
    root = graphframe.graph.roots[0]
    assert flat_profile.loc[root.frame["name"]]["time (inc)"] == np.mean(
        graphframe.dataframe.loc[root]["time (inc)"].to_numpy()
    )

    # Check if the aggregated time is correct for the '__GI_sched_yield' node.
    another = graphframe.dataframe[
        graphframe.dataframe["name"] == "__GI_sched_yield"
    ].index[0][0]
    assert flat_profile.loc[another.frame["name"]]["time (inc)"] == np.mean(
        graphframe.dataframe.loc[another]["time (inc)"].to_numpy()
    )

    # If as_index is False, check if the index is properly numbered
    flat_profile = graphframe.flat_profile(as_index=False)
    assert sorted(flat_profile.index.tolist()) == list(range(len(flat_profile)))


def test_load_imbalance(calc_pi_hpct_db):
    """Validate that the load imbalance is calculated correctly."""

    graphframe = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))
    load_imb_gf = graphframe.load_imbalance(metric_column="time (inc)", threshold=0.01)

    # Check if load imbalance is correct for the root node.
    root = graphframe.graph.roots[0]
    root_metric_max_org = np.max(graphframe.dataframe.loc[root, "time (inc)"])
    root_metric_mean_org = np.mean(graphframe.dataframe.loc[root, "time (inc)"])
    root_imbalance = load_imb_gf.dataframe.loc[root]["time (inc).imbalance"]
    assert root_metric_max_org / root_metric_mean_org == root_imbalance

    # Check if load imbalance is correct for the main node.
    main = root.children[0]
    main_metric_max_org = np.max(graphframe.dataframe.loc[main, "time (inc)"])
    main_metric_mean_org = np.mean(graphframe.dataframe.loc[main, "time (inc)"])
    main_imbalance = load_imb_gf.dataframe.loc[main]["time (inc).imbalance"]
    assert main_metric_max_org / main_metric_mean_org == main_imbalance

    # Check if load imbalance is correct for the the node '__GI_sched_yield'.
    another = graphframe.dataframe[
        graphframe.dataframe["name"] == "__GI_sched_yield"
    ].index[0][0]
    another_lb = load_imb_gf.dataframe[
        load_imb_gf.dataframe["name"] == "__GI_sched_yield"
    ].index[0]
    another_metric_max_org = np.max(graphframe.dataframe.loc[another, "time (inc)"])
    another_metric_mean_org = np.mean(graphframe.dataframe.loc[another, "time (inc)"])
    another_imbalance = load_imb_gf.dataframe.loc[another_lb]["time (inc).imbalance"]
    assert another_metric_max_org / another_metric_mean_org == another_imbalance


def test_hot_path(calc_pi_hpct_db):
    """Validate the hot path with known data from HPCToolkit."""
    graphframe = GraphFrame.from_hpctoolkit(str(calc_pi_hpct_db))

    # Correctness of this list is validated
    # via HPCViewer.
    hpctoolkit_hot_path = [
        "<program root>",
        "main",
        "62:MPI_Finalize",
        "PMPI_Finalize",
        "294:MPID_Finalize",
        "162:MPIDI_CH3_Finalize",
        "230:psm_dofinalize",
        "36:<unknown procedure>",
        "<unknown procedure>",
        "<unknown procedure>",
        "<unknown procedure>",
    ]

    hot_path = graphframe.hot_path(metric="time (inc)")

    for node in hot_path:
        assert node.frame["name"] in hpctoolkit_hot_path

    # test if parameters override.
    # should return the root if threshold equals 2
    hot_path = graphframe.hot_path(metric="time (inc)", threshold=2)
    assert len(hot_path) == 2
    assert hot_path[0].frame["name"] == "<program root>"


def test_multirun_analysis_lulesh(lulesh_caliper_json):
    """Validate that multirun_analysis works correctly with data containing
    non-repeating functions."""
    gf1 = GraphFrame.from_caliper(lulesh_caliper_json)
    gf2, gf4, gf8 = gf1.deepcopy(), gf1.deepcopy(), gf1.deepcopy()

    gf1.update_metadata(1)
    gf2.update_metadata(2)
    gf4.update_metadata(4)
    gf8.update_metadata(8)

    gf1_copy, gf2_copy, gf4_copy, gf8_copy = (
        gf1.deepcopy(),
        gf2.deepcopy(),
        gf4.deepcopy(),
        gf8.deepcopy(),
    )

    # drop index levels, filter values below threshold of 500000, group nodes by name,
    # remove all columns except time, and replace time column with num_processes
    # for each dataframe
    for gf in [gf1, gf2, gf4, gf8]:
        gf.drop_index_levels()
        filtered_rows = gf.dataframe.apply(lambda x: x["time"] > 500000.0, axis=1)
        gf.dataframe = gf.dataframe[filtered_rows]
        gf.dataframe = gf.dataframe.groupby("name").sum()
        gf.dataframe = gf.dataframe[["time"]]
        gf.dataframe = gf.dataframe.rename(
            {"time": gf.metadata["num_processes"]}, axis="columns"
        )

    # join the dataframes column-wise, transpose the dataframe, and set the index name
    # to num_processes
    df_dummy = pd.concat(
        [gf1.dataframe, gf2.dataframe, gf4.dataframe, gf8.dataframe], axis=1
    )
    df_dummy = df_dummy.transpose()
    df_dummy.index.name = "num_processes"

    # run multirun_analyis
    df_test = Chopper.multirun_analysis(
        graphframes=[gf1_copy, gf2_copy, gf4_copy, gf8_copy],
        pivot_index="num_processes",
        columns="name",
        metric="time",
        threshold=500000.0,
    )

    # check if the test and dummy dataframes match
    assert df_test.equals(df_dummy)


def test_multirun_analysis_literal(mock_graph_literal):
    """Validate that multirun_analysis works correctly with data containing
    repeating functions."""
    gf1 = GraphFrame.from_literal(mock_graph_literal)
    gf2, gf4, gf8 = gf1.deepcopy(), gf1.deepcopy(), gf1.deepcopy()

    gf1.update_metadata(1)
    gf2.update_metadata(2)
    gf4.update_metadata(4)
    gf8.update_metadata(8)

    gf1_copy, gf2_copy, gf4_copy, gf8_copy = (
        gf1.deepcopy(),
        gf2.deepcopy(),
        gf4.deepcopy(),
        gf8.deepcopy(),
    )

    # filter values below threshold of 5.0, group nodes by name,
    # remove all columns except time, and replace time column with num_processes
    # for each dataframe
    for gf in [gf1, gf2, gf4, gf8]:
        filtered_rows = gf.dataframe.apply(lambda x: x["time"] > 5.0, axis=1)
        gf.dataframe = gf.dataframe[filtered_rows]
        gf.dataframe = gf.dataframe.drop_duplicates()
        gf.dataframe = gf.dataframe.groupby("name").sum()
        gf.dataframe = gf.dataframe[["time"]]
        gf.dataframe = gf.dataframe.rename(
            {"time": gf.metadata["num_processes"]}, axis="columns"
        )

    # join the dataframes column-wise, transpose the dataframe, and set the index name
    # to num_processes
    df_dummy = pd.concat(
        [gf1.dataframe, gf2.dataframe, gf4.dataframe, gf8.dataframe], axis=1
    )
    df_dummy = df_dummy.transpose()
    df_dummy.index.name = "num_processes"

    # run multirun_analyis
    df_test = Chopper.multirun_analysis(
        graphframes=[gf1_copy, gf2_copy, gf4_copy, gf8_copy],
        pivot_index="num_processes",
        columns="name",
        metric="time",
        threshold=5.0,
    )

    # check if the test and dummy dataframes match
    assert df_test.equals(df_dummy)


def test_speedup_eff_analysis_literal(mock_graph_literal):
    """Validate that speedup_efficiency works correctly."""
    gf1 = GraphFrame.from_literal(mock_graph_literal)

    gf2, gf4, gf8 = gf1.deepcopy(), gf1.deepcopy(), gf1.deepcopy()

    gf2.dataframe["time"] *= 0.5

    gf1.update_metadata(1)
    gf2.update_metadata(2)
    gf4.update_metadata(4)
    gf8.update_metadata(8)

    gf1_copy, gf2_copy, gf4_copy, gf8_copy = (
        gf1.deepcopy(),
        gf2.deepcopy(),
        gf4.deepcopy(),
        gf8.deepcopy(),
    )

    gfs = [gf1_copy, gf2_copy, gf4_copy, gf8_copy]
    eff = Chopper.speedup_efficiency(
        gfs, strong=True, speedup=True, efficiency=False, metrics=["time"]
    )
    assert eff.iloc[1]["2.time.speedup"] == 2.0
    assert eff.iloc[1]["4.time.speedup"] == 1.0

    eff = Chopper.speedup_efficiency(
        gfs, strong=True, speedup=False, efficiency=True, metrics=["time"]
    )
    assert eff.iloc[1]["2.time.efficiency"] == 1.0
    assert eff.iloc[1]["4.time.efficiency"] == 0.25

    eff = Chopper.speedup_efficiency(
        gfs, weak=True, speedup=False, efficiency=True, metrics=["time"]
    )
    assert eff.iloc[1]["2.time.efficiency"] == 2.0
    assert eff.iloc[1]["4.time.efficiency"] == 1.0

    eff = Chopper.speedup_efficiency(
        gfs, strong=True, weak=False, speedup=True, efficiency=True, metrics=["time"]
    )
    assert eff.iloc[1]["2.time.speedup"] == 2.0
    assert eff.iloc[1]["2.time.efficiency"] == 1.0


def test_correlation_analysis_literal(mock_graph_literal):
    """Validate that correlation analysis functions works correctly."""
    gf = GraphFrame.from_literal(mock_graph_literal)

    gf.dataframe["time2"] = gf.dataframe["time"]
    correlation_matrix = gf.correlation_analysis(
        metrics=["time", "time2"], method="spearman"
    )
    assert correlation_matrix.loc["time", "time2"] == 1.0



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet.node import Node
from hatchet.frame import Frame
from hatchet.graph import Graph


def test_from_lists():
    """Ensure we can traverse roots in correct order without repeating a
    shared subdag.
    """
    d = Node(Frame(name="d"))
    diamond_subdag = Node.from_lists(("a", ("b", d), ("c", d)))

    g = Graph.from_lists(("e", "f", diamond_subdag), ("g", diamond_subdag, "h"))
    assert list(g.traverse(attrs="name")) == ["e", "a", "b", "d", "c", "f", "g", "h"]


def test_len_chain():
    graph = Graph.from_lists(("a", "b", "c", "d", "e"))
    assert len(graph) == 5


def test_len_diamond():
    d = Node(Frame(name="d"))
    graph = Graph.from_lists(("a", ("b", d), ("c", d)))
    assert len(graph) == 4


def test_len_tree():
    graph = Graph.from_lists(("a", ("b", "d"), ("c", "d")))
    assert len(graph) == 5


def test_copy():
    d = Node(Frame(name="d"))
    diamond_subdag = Node.from_lists(("a", ("b", d), ("c", d)))
    g = Graph.from_lists(("e", "f", diamond_subdag), ("g", diamond_subdag, "h"))

    assert g.copy() == g


def test_union_dag():
    # make graphs g1, g2, and g3, where you know g3 is the union of g1 and g2
    c = Node.from_lists(("c", "d"))
    g1 = Graph.from_lists(("a", ("b", c), ("e", c, "f")))

    d = Node(Frame(name="d"))
    g2 = Graph.from_lists(("a", ("b", ("c", d)), ("e", d, "f")))

    d2 = Node(Frame(name="d"))
    c2 = Node.from_lists(("c", d2))
    g3 = Graph.from_lists(("a", ("b", c2), ("e", c2, d2, "f")))

    assert g1 != g2

    g4 = g1.union(g2)

    assert g4 == g3


def test_dag_is_not_tree():
    g = Graph.from_lists(("b", "c"), ("d", "e"))
    assert not g.is_tree()

    d = Node(Frame(name="d"))
    diamond_subdag = Node.from_lists(("a", ("b", d), ("c", d)))
    g = Graph([diamond_subdag])
    assert not g.is_tree()

    g = Graph.from_lists(("e", "f", diamond_subdag), ("g", diamond_subdag, "h"))
    assert not g.is_tree()


def test_trees_are_trees():
    g = Graph.from_lists(("a",))
    assert g.is_tree()

    g = Graph.from_lists(("a", ("b", ("c"))))
    assert g.is_tree()

    g = Graph.from_lists(("a", "b", "c"))
    assert g.is_tree()

    g = Graph.from_lists(
        ("a", ("b", "e", "f", "g"), ("c", "e", "f", "g"), ("d", "e", "f", "g"))
    )
    assert g.is_tree()



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import sys
import traceback
import os

from collections import defaultdict

import pandas as pd
import numpy as np
import multiprocess as mp

from .node import Node
from .graph import Graph
from .frame import Frame
from .query import AbstractQuery, QueryMatcher, CypherQuery
from .external.console import ConsoleRenderer
from .util.dot import trees_to_dot
from .util.logger import Logger
from .util.deprecated import deprecated_params
from .chopper import Chopper

try:
    from .cython_modules.libs import graphframe_modules as _gfm_cy
except ImportError:
    print("-" * 80)
    print(
        """Error: Shared object (.so) not found for cython module.\n\tPlease run install.sh from the hatchet root directory to build modules."""
    )
    print("-" * 80)
    traceback.print_exc()
    raise


def parallel_apply(filter_function, dataframe, queue):
    """A function called in parallel, which does a pandas apply on part of a
    dataframe and returns the results via multiprocessing queue function."""
    filtered_rows = dataframe.apply(filter_function, axis=1)
    filtered_df = dataframe[filtered_rows]
    queue.put(filtered_df)


class GraphFrame:
    """An input dataset is read into an object of this type, which includes a graph
    and a dataframe.
    """

    def __init__(
        self,
        graph,
        dataframe,
        exc_metrics,
        inc_metrics,
        default_metric="time",
        metadata=None,
        attributes=None,
    ):
        """Create a new GraphFrame from a graph and a dataframe.

        Likely, you do not want to use this function.

        See ``from_hpctoolkit``, ``from_caliper``, ``from_gprof_dot``, and
        other reader methods for easier ways to create a ``GraphFrame``.

        Arguments:
             graph (Graph): Graph of nodes in this GraphFrame.
             dataframe (DataFrame): Pandas DataFrame indexed by Nodes
                 from the graph, and potentially other indexes.
             exc_metrics: list of names of exclusive metrics in the dataframe.
             inc_metrics: list of names of inclusive metrics in the dataframe.
             default_metric (str): default column to use if one is needed but
                 not explicitly specified
             metadata (dict): information about the dataframe, such as the naming
                 convention for inclusive/exclusive metrics
             attributes (dict): a dictionary of names + values used by readers to
                 extend the Graphframe object. For example, if the underlying reader
                 needs to provide important and easily accessible information or
                 the reader needs to provide an additional function(s) for
                 manipulating/verifying the data.
        """
        if graph is None:
            raise ValueError("GraphFrame() requires a Graph")
        if dataframe is None:
            raise ValueError("GraphFrame() requires a DataFrame")
        if exc_metrics is None and inc_metrics is None:
            raise ValueError(
                "GraphFrame() requires atleast one exclusive or inclusive metric"
            )

        if "node" not in list(dataframe.index.names):
            raise ValueError(
                "DataFrames passed to GraphFrame() must have an index called 'node'."
            )

        self.graph = graph
        self.dataframe = dataframe
        self.exc_metrics = [] if exc_metrics is None else exc_metrics
        self.inc_metrics = [] if inc_metrics is None else inc_metrics
        self.default_metric = default_metric
        self.metadata = {} if metadata is None else metadata
        if "hatchet_inclusive_suffix" not in self.metadata:
            self.metadata["hatchet_inclusive_suffix"] = " (inc)"
        if "hatchet_exclusive_suffix" not in self.metadata:
            self.metadata["hatchet_exclusive_suffix"] = " (exc)"

        try:
            attributes = dict(attributes) if attributes is not None else {}
        except TypeError as e:
            print(
                "graphframe attributes argument must be convertable to dict: {}".format(
                    e
                )
            )
            raise

        # create attributes
        self.attributes = attributes.keys()
        for x, y in attributes.items():
            if hasattr(self, x):
                raise ValueError(
                    "cannot create graphframe attribute '{}' because it already exists".format(
                        x
                    )
                )
            setattr(self, x, y)

    @staticmethod
    @Logger.loggable
    def from_hpctoolkit(dirname):
        """Read an HPCToolkit database directory into a new GraphFrame.

        Arguments:
            dirname (str): parent directory of an HPCToolkit
                experiment.xml file

        Returns:
            (GraphFrame): new GraphFrame containing HPCToolkit profile data
        """
        # import this lazily to avoid circular dependencies
        from .readers.hpctoolkit_reader import HPCToolkitReader
        from .readers.hpctoolkit_v4_reader import HPCToolkitV4Reader

        if "experiment.xml" in os.listdir(dirname):
            return HPCToolkitReader(dirname).read()
        else:
            return HPCToolkitV4Reader(dirname).read()

    @staticmethod
    @Logger.loggable
    def from_caliper(filename_or_stream, query=None):
        """Read in a Caliper .cali or .json file.
        Args:
            filename_or_stream (str or file-like): name of a Caliper output
                file in `.cali` or JSON-split format, or an open file object
                to read one
            query (str): cali-query in CalQL format
        """
        # import this lazily to avoid circular dependencies
        from .readers.caliper_reader import CaliperReader

        return CaliperReader(filename_or_stream, query).read()

    @staticmethod
    @Logger.loggable
    def from_caliperreader(filename_or_caliperreader):
        """Read in a native Caliper `cali' file using Caliper's python reader.

        Args:
            filename_or_caliperreader (str or CaliperReader): name of a Caliper
                output file in `.cali` format, or a CaliperReader object
        """
        # import this lazily to avoid circular dependencies
        from .readers.caliper_native_reader import CaliperNativeReader

        return CaliperNativeReader(filename_or_caliperreader).read()

    @staticmethod
    @Logger.loggable
    def from_spotdb(db_key, list_of_ids=None):
        """Read multiple graph frames from a SpotDB instance

        Args:
            db_key (str or SpotDB object): locator for SpotDB instance
                This can be a SpotDB object directly, or a locator for a spot
                database, which is a string with either
                    * A directory for .cali files,
                    * A .sqlite file name
                    * A SQL database URL (e.g., "mysql://hostname/db")

            list_of_ids: The list of run IDs to read from the database.
                If this is None, returns all runs.

        Returns:
            A list of graphframes, one for each requested run that was found
        """

        from .readers.spotdb_reader import SpotDBReader

        return SpotDBReader(db_key, list_of_ids).read()

    @staticmethod
    @Logger.loggable
    def from_gprof_dot(filename):
        """Read in a DOT file generated by gprof2dot."""
        # import this lazily to avoid circular dependencies
        from .readers.gprof_dot_reader import GprofDotReader

        return GprofDotReader(filename).read()

    @staticmethod
    @Logger.loggable
    def from_cprofile(filename):
        """Read in a pstats/prof file generated using python's cProfile."""
        # import this lazily to avoid circular dependencies
        from .readers.cprofile_reader import CProfileReader

        return CProfileReader(filename).read()

    @staticmethod
    @Logger.loggable
    def from_pyinstrument(filename):
        """Read in a JSON file generated using Pyinstrument."""
        # import this lazily to avoid circular dependencies
        from .readers.pyinstrument_reader import PyinstrumentReader

        return PyinstrumentReader(filename).read()

    @staticmethod
    @Logger.loggable
    def from_tau(dirname):
        """Read in a profile generated using TAU."""
        # import this lazily to avoid circular dependencies
        from .readers.tau_reader import TAUReader

        return TAUReader(dirname).read()

    @staticmethod
    @Logger.loggable
    def from_scorep(filename):
        """Read in a profile generated using Score-P."""
        # import this lazily to avoid circular dependencies
        from .readers.scorep_reader import ScorePReader

        return ScorePReader(filename).read()

    @staticmethod
    @Logger.loggable
    def from_timemory(input=None, select=None, **_kwargs):
        """Read in timemory data.

        Links:
            https://github.com/NERSC/timemory
            https://timemory.readthedocs.io

        Arguments:
            input (str or file-stream or dict or None):
                Valid argument types are:

                1. Filename for a timemory JSON tree file
                2. Open file stream to one of these files
                3. Dictionary from timemory JSON tree


                Currently, timemory supports two JSON layouts: flat and tree.
                The former is a 1D-array representation of the hierarchy which
                represents the hierarchy via indentation schemes in the labels
                and is not compatible with hatchet. The latter is a hierarchical
                representation of the data and is the required JSON layout when
                using hatchet. Timemory JSON tree files typically have the
                extension ".tree.json".

                If input is None, this assumes that timemory has been recording
                data within the application that is using hatchet. In this
                situation, this method will attempt to import the data directly
                from timemory.

                At the time of this writing, the direct data import will:

                1. Stop any currently collecting components
                2. Aggregate child thread data of the calling thread
                3. Clear all data on the child threads
                4. Aggregate the data from any MPI and/or UPC++ ranks.


                Thus, if MPI or UPC++ is used, every rank must call this routine.
                The zeroth rank will have the aggregation and all the other
                non-zero ranks will only have the rank-specific data.

                Whether or not the per-thread and per-rank data itself is
                combined is controlled by the `collapse_threads` and
                `collapse_processes` attributes in the `timemory.settings`
                submodule.

                In the C++ API, it is possible for only #1 to be applied and data
                can be obtained for an individual thread and/or rank without
                aggregation. This is not currently available to Python, however,
                it can be made available upon request via a GitHub Issue.

            select (list of str):
                A list of strings which match the component enumeration names, e.g. ["cpu_clock"].

            per_thread (boolean):
                Ensures that when applying filters to the graphframe, frames with
                identical name/file/line/etc. info but from different threads are
                not combined

            per_rank (boolean):
                Ensures that when applying filters to the graphframe, frames with
                identical name/file/line/etc. info but from different ranks are
                not combined

        """
        from .readers.timemory_reader import TimemoryReader

        if input is not None:
            try:
                return TimemoryReader(input, select, **_kwargs).read()
            except IOError:
                pass
        else:
            try:
                import timemory

                TimemoryReader(timemory.get(hierarchy=True), select, **_kwargs).read()
            except ImportError:
                print(
                    "Error! timemory could not be imported. Provide filename, file stream, or dict."
                )
                raise

    @staticmethod
    @Logger.loggable
    def from_literal(graph_dict):
        """Create a GraphFrame from a list of dictionaries."""
        # import this lazily to avoid circular dependencies
        from .readers.literal_reader import LiteralReader

        return LiteralReader(graph_dict).read()

    @staticmethod
    @Logger.loggable
    def from_apex(dirname):
        """Create a GraphFrame from a list of dictionaries."""
        # import this lazily to avoid circular dependencies
        from .readers.apex_reader import ApexReader

        return ApexReader(dirname).read()

    @staticmethod
    @Logger.loggable
    def from_perfetto(file_or_files, **kwargs):
        """Create a GraphFrame from perfetto files"""
        # import this lazily to avoid circular dependencies
        from .readers.perfetto_reader import PerfettoReader
        import glob
        import os

        files = []
        filename_list = (
            file_or_files
            if isinstance(file_or_files, (list, tuple))
            else [file_or_files]
        )

        for filename in filename_list:
            if os.path.exists(filename) and os.path.isdir(filename):
                filename = os.path.join(filename, "*.proto")

            if not os.path.exists(filename):
                files += glob.glob(
                    filename,
                    recursive=kwargs["recursive"] if "recursive" in kwargs else False,
                )
            else:
                files += [filename]

        if len(files) == 0:
            raise ValueError(
                "No omnitrace perfetto files found in '{}'".format(filename_list)
            )

        return PerfettoReader(files, **kwargs).read(**kwargs)

    @staticmethod
    @Logger.loggable
    def from_omnitrace(file_or_files, **kwargs):
        """In the future, this should support both omnitrace perfetto traces and timemory JSON files"""
        return GraphFrame.from_perfetto(file_or_files, **kwargs)

    @staticmethod
    @Logger.loggable
    def from_lists(*lists):
        """Make a simple GraphFrame from lists.

        This creates a Graph from lists (see ``Graph.from_lists()``) and uses
        it as the index for a new GraphFrame. Every node in the new graph has
        exclusive time of 1 and inclusive time is computed automatically.

        """
        graph = Graph.from_lists(*lists)
        graph.enumerate_traverse()

        df = pd.DataFrame({"node": list(graph.traverse())})
        df["time"] = [1.0] * len(graph)
        df["name"] = [n.frame["name"] for n in graph.traverse()]
        df.set_index(["node"], inplace=True)
        df.sort_index(inplace=True)

        gf = GraphFrame(graph, df, ["time"], [])
        gf.calculate_inclusive_metrics()
        return gf

    @staticmethod
    @Logger.loggable
    def from_hdf(filename, **kwargs):
        # import this lazily to avoid circular dependencies
        from .readers.hdf5_reader import HDF5Reader

        return HDF5Reader(filename).read(**kwargs)

    @Logger.loggable
    def to_hdf(self, filename, key="hatchet_graphframe", **kwargs):
        # import this lazily to avoid circular dependencies
        from .writers.hdf5_writer import HDF5Writer

        HDF5Writer(filename).write(self, key=key, **kwargs)

    @Logger.loggable
    def update_metadata(self, num_processes=None, num_threads=None, metadata=None):
        """Update a GraphFrame object's metadata."""
        if num_processes is not None:
            if not isinstance(num_processes, int):
                raise TypeError(
                    "The number of processing elements must be of type integer."
                )
            self.metadata["num_processes"] = num_processes

        if num_threads is not None:
            if not isinstance(num_threads, int):
                raise TypeError("The number of threads must be of type integer.")
            self.metadata["num_threads"] = num_threads

        if metadata is not None:
            self.metadata.update(metadata)

        return self

    @Logger.loggable
    def copy(self):
        """Return a shallow copy of the graphframe.

        This copies the DataFrame, but the Graph is shared between self and
        the new GraphFrame.
        """
        return GraphFrame(
            self.graph,
            self.dataframe.copy(),
            list(self.exc_metrics),
            list(self.inc_metrics),
            self.default_metric,
            dict(self.metadata),
            attributes=dict([[x, getattr(self, x)] for x in self.attributes]),
        )

    @Logger.loggable
    def deepcopy(self):
        """Return a copy of the graphframe."""
        node_clone = {}
        graph_copy = self.graph.copy(node_clone)
        dataframe_copy = self.dataframe.copy()

        index_names = dataframe_copy.index.names
        dataframe_copy.reset_index(inplace=True)

        dataframe_copy["node"] = dataframe_copy["node"].apply(lambda x: node_clone[x])

        dataframe_copy.set_index(index_names, inplace=True)

        return GraphFrame(
            graph_copy,
            dataframe_copy,
            list(self.exc_metrics),
            list(self.inc_metrics),
            self.default_metric,
            dict(self.metadata),
            attributes=dict([[x, getattr(self, x)] for x in self.attributes]),
        )

    def drop_index_levels(self, function=np.mean):
        """Drop all index levels but `node`."""
        index_names = list(self.dataframe.index.names)
        index_names.remove("node")

        # create dict that stores aggregation function for each column
        agg_dict = {}
        for col in self.dataframe.columns.tolist():
            if col in self.exc_metrics + self.inc_metrics:
                agg_dict[col] = function
            else:
                agg_dict[col] = lambda x: x.iloc[0]

        # perform a groupby to merge nodes that just differ in index columns
        self.dataframe.reset_index(level="node", inplace=True)
        agg_df = self.dataframe.groupby("node").agg(agg_dict)

        self.dataframe = agg_df

    @Logger.loggable
    def filter(self, filter_obj, squash=True, num_procs=mp.cpu_count()):
        """Filter the dataframe using a user-supplied function.

        Note: Operates in parallel on user-supplied lambda functions.

        Arguments:
            filter_obj (callable, list, or QueryMatcher): the filter to apply to the GraphFrame.
            squash (boolean, optional): if True, automatically call squash for the user.
        """
        dataframe_copy = self.dataframe.copy()

        index_names = self.dataframe.index.names
        dataframe_copy.reset_index(inplace=True)

        filtered_df = None

        if callable(filter_obj):
            # applying pandas filter using the callable function
            if num_procs > 1:
                # perform filter in parallel (default)
                queue = mp.Queue()
                processes = []
                returned_frames = []
                subframes = np.array_split(dataframe_copy, num_procs)

                # Manually create a number of processes equal to the number of
                # logical cpus available
                for pid in range(num_procs):
                    process = mp.Process(
                        target=parallel_apply,
                        args=(filter_obj, subframes[pid], queue),
                    )
                    process.start()
                    processes.append(process)

                # Stores filtered subframes in a list: 'returned_frames', for
                # pandas concatenation. This intermediary list is used because
                # pandas concat is faster when called only once on a list of
                # dataframes, than when called multiple times appending onto a
                # frame of increasing size.
                for pid in range(num_procs):
                    returned_frames.append(queue.get())

                for proc in processes:
                    proc.join()

                filtered_df = pd.concat(returned_frames)

            else:
                # perform filter sequentiually if num_procs = 1
                filtered_rows = dataframe_copy.apply(filter_obj, axis=1)
                filtered_df = dataframe_copy[filtered_rows]

        elif isinstance(filter_obj, (list, str)) or issubclass(
            type(filter_obj), AbstractQuery
        ):
            # use a callpath query to apply the filter
            query = filter_obj
            if isinstance(filter_obj, list):
                query = QueryMatcher(filter_obj)
            elif isinstance(filter_obj, str):
                query = CypherQuery(filter_obj)
            query_matches = query.apply(self)
            # match_set = list(set().union(*query_matches))
            # filtered_df = dataframe_copy.loc[dataframe_copy["node"].isin(match_set)]
            filtered_df = dataframe_copy.loc[dataframe_copy["node"].isin(query_matches)]
        else:
            raise InvalidFilter(
                "The argument passed to filter must be a callable, a query path list, or a QueryMatcher object."
            )

        if filtered_df.shape[0] == 0:
            raise EmptyFilter(
                "The provided filter would have produced an empty GraphFrame."
            )

        filtered_df.set_index(index_names, inplace=True)

        filtered_gf = GraphFrame(
            self.graph,
            filtered_df,
            list(self.exc_metrics),
            list(self.inc_metrics),
            self.default_metric,
            dict(self.metadata),
            attributes=dict([[x, getattr(self, x)] for x in self.attributes]),
        )

        if squash:
            return filtered_gf.squash()
        return filtered_gf

    @Logger.loggable
    def squash(self):
        """Rewrite the Graph to include only nodes present in the DataFrame's rows.

        This can be used to simplify the Graph, or to normalize Graph
        indexes between two GraphFrames.
        """
        index_names = self.dataframe.index.names
        self.dataframe.reset_index(inplace=True)

        # create new nodes for each unique node in the old dataframe
        old_to_new = {n: n.copy() for n in set(self.dataframe["node"])}
        for i in old_to_new:
            old_to_new[i]._hatchet_nid = i._hatchet_nid

        # Maintain sets of connections to make for each old node.
        # Start with old -> new mapping and update as we traverse subgraphs.
        connections = defaultdict(lambda: set())
        connections.update({k: {v} for k, v in old_to_new.items()})

        new_roots = []  # list of new roots

        # connect new nodes to children according to transitive
        # relationships in the old graph.
        def rewire(node, new_parent, visited):
            # make all transitive connections for the node we're visiting
            for n in connections[node]:
                if new_parent:
                    # there is a parent in the new graph; connect it
                    if n not in new_parent.children:
                        new_parent.add_child(n)
                        n.add_parent(new_parent)

                elif n not in new_roots:
                    # this is a new root
                    new_roots.append(n)

            new_node = old_to_new.get(node)
            transitive = set()
            if node not in visited:
                visited.add(node)
                for child in node.children:
                    transitive |= rewire(child, new_node or new_parent, visited)

            if new_node:
                # since new_node exists in the squashed graph, we only
                # need to connect new_node
                return {new_node}
            else:
                # connect parents to the first transitively reachable
                # new_nodes of nodes we're removing with this squash
                connections[node] |= transitive
                return connections[node]

        # run rewire for each root and make a new graph
        visited = set()
        for root in self.graph.roots:
            rewire(root, None, visited)
        graph = Graph(new_roots)
        graph.enumerate_traverse()

        # reindex new dataframe with new nodes
        df = self.dataframe.copy()
        df["node"] = df["node"].apply(lambda x: old_to_new[x])

        # at this point, the graph is potentially invalid, as some nodes
        # may have children with identical frames.
        merges = graph.normalize()
        df["node"] = df["node"].apply(lambda n: merges.get(n, n))

        self.dataframe.set_index(index_names, inplace=True)
        df.set_index(index_names, inplace=True)
        # create dict that stores aggregation function for each column
        agg_dict = {}
        for col in df.columns.tolist():
            if col in self.exc_metrics + self.inc_metrics:
                # use min_count=1 (default is 0) here, so sum of an all-NA
                # series is NaN, not 0
                # when min_count=1, sum([NaN, NaN)] = NaN
                # when min_count=0, sum([NaN, NaN)] = 0
                agg_dict[col] = lambda x: x.sum(min_count=1)
            else:
                agg_dict[col] = lambda x: x.iloc[0]

        # perform a groupby to merge nodes with the same callpath
        agg_df = df.groupby(index_names).agg(agg_dict)
        agg_df.sort_index(inplace=True)

        # put it all together
        new_gf = GraphFrame(
            graph,
            agg_df,
            list(self.exc_metrics),
            list(self.inc_metrics),
            self.default_metric,
            dict(self.metadata),
            attributes=dict([[x, getattr(self, x)] for x in self.attributes]),
        )
        new_gf.calculate_inclusive_metrics()
        return new_gf

    def _init_sum_columns(self, columns, out_columns):
        """Helper function for subtree_sum and subgraph_sum."""
        if out_columns is None:
            out_columns = columns
        else:
            # init out columns with input columns in case they are not there.
            for col, out in zip(columns, out_columns):
                self.dataframe[out] = self.dataframe[col]

        if len(columns) != len(out_columns):
            raise ValueError("columns out_columns must be the same length!")

        return out_columns

    def subtree_sum(
        self, columns, out_columns=None, function=lambda x: x.sum(min_count=1)
    ):
        """Compute sum of elements in subtrees.  Valid only for trees.

        For each row in the graph, ``out_columns`` will contain the
        element-wise sum of all values in ``columns`` for that row's node
        and all of its descendants.

        This algorithm will multiply count nodes with in-degree higher
        than one -- i.e., it is only correct for trees.  Prefer using
        ``subgraph_sum`` (which calls ``subtree_sum`` if it can), unless
        you have a good reason not to.

        Arguments:
            columns (list of str): names of columns to sum (default: all columns)
            out_columns (list of str): names of columns to store results
                (default: in place)
            function (callable): associative operator used to sum
                elements, sum of an all-NA series is NaN (default: sum(min_count=1))
        """
        out_columns = self._init_sum_columns(columns, out_columns)

        # sum over the output columns
        for node in self.graph.traverse(order="post"):
            if node.children:
                # TODO: need a better way of aggregating inclusive metrics when
                # TODO: there is a multi-index
                try:
                    is_multi_index = isinstance(
                        self.dataframe.index, pd.core.index.MultiIndex
                    )
                except AttributeError:
                    is_multi_index = isinstance(self.dataframe.index, pd.MultiIndex)

                if is_multi_index:
                    for rank_thread in self.dataframe.loc[
                        (node), out_columns
                    ].index.unique():
                        # rank_thread is either rank or a tuple of (rank, thread).
                        # We check if rank_thread is a tuple and if it is, we
                        # create a tuple of (node, rank, thread). If not, we create
                        # a tuple of (node, rank).
                        if isinstance(rank_thread, tuple):
                            df_index1 = (node,) + rank_thread
                            df_index2 = ([node] + node.children,) + rank_thread
                        else:
                            df_index1 = (node, rank_thread)
                            df_index2 = ([node] + node.children, rank_thread)

                        for col in out_columns:
                            self.dataframe.loc[df_index1, col] = function(
                                self.dataframe.loc[df_index2, col]
                            )
                else:
                    for col in out_columns:
                        self.dataframe.loc[node, col] = function(
                            self.dataframe.loc[[node] + node.children, col]
                        )

    def subgraph_sum(
        self, columns, out_columns=None, function=lambda x: x.sum(min_count=1)
    ):
        """Compute sum of elements in subgraphs.

        For each row in the graph, ``out_columns`` will contain the
        element-wise sum of all values in ``columns`` for that row's node
        and all of its descendants.

        This algorithm is worst-case quadratic in the size of the graph,
        so we try to call ``subtree_sum`` if we can.  In general, there
        is not a particularly efficient algorithm known for subgraph
        sums, so this does about as well as we know how.

        Arguments:
            columns (list of str):  names of columns to sum (default: all columns)
            out_columns (list of str): names of columns to store results
                (default: in place)
            function (callable): associative operator used to sum
                elements, sum of an all-NA series is NaN (default: sum(min_count=1))
        """
        if self.graph.is_tree():
            self.subtree_sum(columns, out_columns, function)
            return

        out_columns = self._init_sum_columns(columns, out_columns)
        for node in self.graph.traverse():
            subgraph_nodes = list(node.traverse())
            # TODO: need a better way of aggregating inclusive metrics when
            # TODO: there is a multi-index
            try:
                is_multi_index = isinstance(
                    self.dataframe.index, pd.core.index.MultiIndex
                )
            except AttributeError:
                is_multi_index = isinstance(self.dataframe.index, pd.MultiIndex)

            if is_multi_index:
                for rank_thread in self.dataframe.loc[
                    (node), out_columns
                ].index.unique():
                    # rank_thread is either rank or a tuple of (rank, thread).
                    # We check if rank_thread is a tuple and if it is, we
                    # create a tuple of (node, rank, thread). If not, we create
                    # a tuple of (node, rank).
                    if isinstance(rank_thread, tuple):
                        df_index1 = (node,) + rank_thread
                        df_index2 = (subgraph_nodes,) + rank_thread
                    else:
                        df_index1 = (node, rank_thread)
                        df_index2 = (subgraph_nodes, rank_thread)

                    for col in out_columns:
                        self.dataframe.loc[df_index1, col] = [
                            function(self.dataframe.loc[df_index2, col])
                        ]
            else:
                # TODO: if you take the list constructor away from the
                # TODO: assignment below, this assignment gives NaNs. Why?
                self.dataframe.loc[(node), out_columns] = list(
                    function(self.dataframe.loc[(subgraph_nodes), columns])
                )

    def calculate_inclusive_metrics(self):
        """Update inclusive columns (typically after operations that rewire the
        graph.
        """
        # we should update inc metric only if exc metric exist
        if not self.exc_metrics:
            return

        self.inc_metrics = [
            "%s%s" % (s, self.metadata["hatchet_inclusive_suffix"])
            for s in self.exc_metrics
        ]
        self.subgraph_sum(self.exc_metrics, self.inc_metrics)

    @Logger.loggable
    def calculate_exclusive_metrics(self, columns=None):
        """Calculates exclusive metrics using the corresponding inclusive metrics.

        Computed for all inclusive metrics if columns=None.

        Raise error if the given column is not in inc_metrics.
        If the given column is inclusive:
            If ' (inc)' is in the given column, name of the new column will be
        the given column without ' (inc)' at the end.
            If ' (inc)' is not in the given column, name of the new column will be
        the given column with ' (exc)' at the end.
        """

        # check if the columns parameter is None
        if columns is not None:
            # make it a list if it's not None and not a list
            if not isinstance(columns, list):
                columns = [columns]

            # check if the given columns are in inc_metrics.
            for column in columns:
                assert (
                    column in self.inc_metrics
                ), "{} does not exist in the graphframe.inc_metrics.".format(column)
        else:
            # if the user doesn't provide any columns, calculate
            # exclusive metric for all inclusive metrics.
            columns = [column for column in self.inc_metrics]

        # create the new columns and add
        # them to the exc_metrics.
        inc_exc_pairs = []
        new_data = {}
        for column in columns:
            # name the new column removing 'hatchet_inc_suffix' from
            # the inclusive column
            new_column = ""
            if self.metadata["hatchet_inclusive_suffix"] in column:
                new_column = column.replace(
                    self.metadata["hatchet_inclusive_suffix"], ""
                )
            # add 'hatchet_ex_suffix' if 'hatchet_inc_suffix' doesn't exists.
            else:
                new_column = column + self.metadata["hatchet_exclusive_suffix"]

            if new_column not in self.exc_metrics:
                self.exc_metrics.append(new_column)

            # keep the columns as a list of (inc_metric, exc_metric)
            inc_exc_pairs.append((column, new_column))
            # create a dict for the new data
            new_data[new_column] = {}

        # compute exclusive metrics by travering the graph
        for node in self.graph.traverse():
            for inc, exc in inc_exc_pairs:
                # if the dataframe index is a MultiIndex
                if isinstance(self.dataframe.index, pd.MultiIndex):
                    for idx in self.dataframe.loc[node].index:
                        node_index_tuple = None
                        # if 1D multiindex
                        if isinstance(idx, int):
                            idx = [idx]
                        node_index_tuple = tuple([node]) + tuple(idx)

                        # calculate exclusive metric.
                        child_inc_sum = 0
                        for child in node.children:
                            child_index_tuple = tuple([child]) + tuple(idx)
                            child_inc_sum += self.dataframe.loc[child_index_tuple][inc]
                        exc_value = (
                            self.dataframe.loc[node_index_tuple][inc] - child_inc_sum
                        )
                        # store the new value.
                        new_data[exc][node_index_tuple] = exc_value
                # if not a MultiIndex
                else:
                    # calculate exclusive metric.
                    child_inc_sum = 0
                    for child in node.children:
                        child_inc_sum += self.dataframe.loc[child][inc]
                    exc_value = self.dataframe.loc[node][inc] - child_inc_sum
                    # store the new value.
                    new_data[exc][node] = exc_value

                # create series for each exc column.
                new_data[exc] = pd.Series(data=new_data[exc])

        # add all new exc columns to the dataframe at once.
        self.dataframe = self.dataframe.assign(**new_data)

    @Logger.loggable
    def show_metric_columns(self):
        """Returns a list of dataframe column labels."""
        return list(self.exc_metrics + self.inc_metrics)

    @Logger.loggable
    def groupby_callpath(self, callpath_to_node_dicts=None):
        """ "Merges the callpaths in a graphframe when the callpaths
        are exactly the same. Returns a new graphframe.
        'callpath_to_node_dicts' parameter is passed by reference
        so updated version of it can be used after calling this function.

        Merging is done using bottom-up approach. It merges starting
        from the longest callpath (the bottom) and goes upward.
        """

        def _merge_paths(
            new_parent, callpath, same_paths, same_paths_dicts, callpath_to_node_dicts
        ):
            node_dicts = same_paths_dicts[callpath]

            new_node = node_dicts[0]["node"].copy()
            # print("NEW NODE:", new_node)
            # print("NODE DICTS:", node_dicts)

            new_node_dict = dict()
            for key, value in node_dicts[0].items():
                new_node_dict[key] = value
            new_node_dict["node"] = new_node

            # since we copied the node_dict (metric values) of
            # the first duplicate node, we add the metric values
            # of the second node.
            for node_dict in node_dicts[1:]:
                for metric in graphframe_cp.inc_metrics + graphframe_cp.exc_metrics:
                    # aggregate the metric values.
                    new_node_dict[metric] += node_dict[metric]

            # set the parent-child relationships for the
            # new node.
            for node_dict in node_dicts:
                # print("show old nodes children:", node_dict["node"].children)
                # add the children of the nodes that have
                # the same callpath to the new node.
                if node_dict["node"].children:
                    for child in node_dict["node"].children:
                        new_node.add_child(child)
                        child.add_parent(new_node)
                        child.parents.remove(node_dict["node"])
                        if new_parent:
                            node_dict["node"].parents = []

                    node_dict["node"].children = []

            parents_children = []
            if new_parent:
                parents_children = []
                for child in new_parent.children:
                    if child.__str__() != new_node.__str__():
                        parents_children.append(child)
                new_parent.children = parents_children
                new_node.add_parent(new_parent)
                new_parent.add_child(new_node)
            else:
                old_parent = node_dicts[0]["node"].parents[0]
                new_node.add_parent(old_parent)
                for child in old_parent.children:
                    if child.__str__() != new_node.__str__():
                        parents_children.append(child)
                old_parent.children = parents_children
                node_dicts[0]["node"].parents[0].add_child(new_node)

            # store new node dict after doing aggregation and
            # setting the relationships.
            del callpath_to_node_dicts[callpath]
            callpath_to_node_dicts[callpath] = [new_node_dict]

            same_paths.remove(callpath)
            for child in new_node.children:
                child_path = child.path()
                if child_path:
                    child_path = child.convert_path_to_str(child_path)
                    if child_path in same_paths:
                        _merge_paths(
                            new_node,
                            child_path,
                            same_paths,
                            same_paths_dicts,
                            callpath_to_node_dicts,
                        )

        if callpath_to_node_dicts is None:
            callpath_to_node_dicts = {}

        graphframe_cp = self.deepcopy()
        graphframe_cp.drop_index_levels(np.max)

        # traverse the graph
        for node in graphframe_cp.graph.traverse():
            # check each path
            for path in node.paths():
                if path is not None:
                    # transform path from tuple of node objects
                    # to tuple of strings.
                    callpath = node.convert_path_to_str(path)

                    # get the metric information from each row
                    # and store them in a dict.
                    node_dict = dict()
                    for metric, metric_value in (
                        graphframe_cp.dataframe.loc[node].to_dict().items()
                    ):
                        node_dict[metric] = metric_value
                    # metric info doesn't contain node since it's index,
                    # so we add the node to the node dict.
                    node_dict["node"] = node

                    # if the callpath was seen before, append the
                    # new node dict. this will be used later for
                    # aggregation
                    if callpath in callpath_to_node_dicts.keys():
                        previous = callpath_to_node_dicts[callpath]
                        previous.append(node_dict)
                        callpath_to_node_dicts[callpath] = previous
                    # if the callpath was not seen before, keep the
                    # related node dict
                    else:
                        callpath_to_node_dicts[callpath] = dict()
                        callpath_to_node_dicts[callpath] = [node_dict]

        same_paths_dicts = {}
        same_paths = []
        # keep the same paths and only work on them.
        for callpath, node_dicts in callpath_to_node_dicts.items():
            # only the duplicate callpaths are stored as list.
            if len(node_dicts) > 1:
                same_paths_dicts[callpath] = node_dicts
                same_paths.append(callpath)

        # sort the duplicate callpaths.
        # we merge them starting from the longest callpath,
        # means that we start from the bottom.
        same_paths.sort(reverse=True)
        while same_paths:
            _merge_paths(
                None,
                same_paths[-1],
                same_paths,
                same_paths_dicts,
                callpath_to_node_dicts,
            )

        # create a new graph by using the updated
        # relationships.
        roots = graphframe_cp.graph.roots
        graph = Graph(roots)
        graph.enumerate_traverse()

        # create a new dataframe that has the aggregate values
        # for the duplicate callpaths.
        all_info = []
        for val in callpath_to_node_dicts.values():
            all_info.append(val[0])
        dataframe = pd.DataFrame(all_info)
        dataframe.set_index("node", inplace=True)
        dataframe.sort_index(inplace=True)

        # create a new graphframe.
        graphframe_new = GraphFrame(
            graph,
            dataframe,
            graphframe_cp.exc_metrics,
            graphframe_cp.inc_metrics,
            graphframe_cp.default_metric,
        )

        graphframe_new.metadata = self.metadata

        return graphframe_new

    @staticmethod
    @Logger.loggable
    def unify_multiple_graphframes(graphframes, num_procs="num_processes"):
        """Unifies multiple graphframes.
        Follows these main steps:
          1. Finds the biggest graphframe (i.e. the graphframe that has more
          indices in the dataframe).
          2. Merges all the nodes that have the same callpath in the biggest
          graphframe.
          3. For each node in the biggest graphframe, checks if other
          graphframes contains some nodes that have the same callpath.
          4. If yes, adds metrics from the other graphframes.
          5. If no, create the corresponding nodes.

        Updates graphframes in place. The graphframes use the same graph with
        their updated dataframes."""

        def _rename_colums(graphframe, metrics, suffix, add):
            "Renames columns and updates inc/exc columns lists."
            if add:
                for idx in range(len(metrics)):
                    graphframe.dataframe.rename(
                        columns={metrics[idx]: "{}-{}".format(metrics[idx], suffix)},
                        inplace=True,
                    )

                    # update inclusive and exclusive metrics lists.
                    metrics[idx] = "{}-{}".format(metrics[idx], str(suffix))
            else:
                for idx in range(len(metrics)):
                    renamed_metric = metrics[idx].replace("-{}".format(suffix), "")
                    graphframe.dataframe.rename(
                        columns={metrics[idx]: renamed_metric},
                        inplace=True,
                    )
                    metrics[idx] = renamed_metric

        def _create_node(biggest_gf, graphframe, node, callpath, callpath_to_node_dict):
            """Recursively creates the missing nodes from bottom to up."""
            # copy and create the new node. it will be added
            # to the 'biggest graphframe'.
            new_node = node.copy()
            # convert callpaths from the tuple of node objects to
            # tuple of strings.
            callpath_names = node.convert_path_to_str(callpath)
            # convert parent's callpath.
            parent_callpath = callpath[:-1]
            parent_callpath_names = node.convert_path_to_str(parent_callpath)

            # keep as root if the node has no parent.
            if not parent_callpath:
                biggest_gf.graph.roots.append(new_node)
                return new_node

            # if parent is found, add child-parent relationships and
            # create a new node dict to store metric values.
            if parent_callpath_names in callpath_to_node_dict.keys():
                # store metric values.
                node_dict = graphframe.dataframe.loc[node].to_dict()
                node_dict["node"] = new_node
                callpath_to_node_dict[callpath_names] = [node_dict]

                # set parent-child relationships.
                parent = callpath_to_node_dict[parent_callpath_names][0]["node"]
                parent.add_child(new_node)
                new_node.add_parent(parent)
                return new_node

            # if parent is not found, recursively go up on the tree
            # by creating parents until finding one.
            parent = _create_node(
                biggest_gf,
                graphframe,
                parent_callpath[-1],
                parent_callpath,
                callpath_to_node_dict,
            )
            # create the new node and set its parent-child
            # relationships.
            node_dict = graphframe.dataframe.loc[node].to_dict()
            node_dict["node"] = new_node
            callpath_to_node_dict[callpath_names] = node_dict

            parent.add_child(new_node)
            new_node.add_parent(parent)
            return new_node

        # check if a list of graphframes
        # is given. if not, make it a list.
        if not isinstance(graphframes, list):
            graphframes = list(graphframes)
        # chose the biggest graphframe and start with
        # it. no need to recreate nodes and relationships
        # for the biggest gf.
        max_num_of_idx = 0
        biggest_gf = None
        gf_to_visited_node = {}
        for idx in range(len(graphframes)):
            # do not change the given gfs.
            # gf_copy = graphframes[idx].deepcopy()
            graphframes[idx].drop_index_levels(np.max)
            gf_to_visited_node[graphframes[idx]] = []

            # rename inc/exc metrics. update dataframe, default metric,
            # and inc/exc lists.
            # old_default_metric = graphframes[idx].default_metric
            assert (
                num_procs in graphframes[idx].metadata.keys()
            ), "{} missing from GraphFrame metadata: use update_metadata() to specify.".format(
                num_procs
            )
            _rename_colums(
                graphframes[idx],
                graphframes[idx].inc_metrics,
                graphframes[idx].metadata[num_procs],
                add=True,
            )
            _rename_colums(
                graphframes[idx],
                graphframes[idx].exc_metrics,
                graphframes[idx].metadata[num_procs],
                add=True,
            )

            # find the biggest graphframe by using the number
            # of indices.
            size_of_df = len(graphframes[idx].dataframe.index)
            if size_of_df > max_num_of_idx:
                biggest_gf = graphframes[idx]
                max_num_of_idx = size_of_df
        # keep the biggest graphframe separate.
        gf_to_visited_node.pop(biggest_gf)
        # merge the nodes that have the exact same callpaths.
        # now we have a graphframe in which all the nodes have
        # unique callpaths.
        # 'callpath_to_node_biggest' dict created outside because
        # we want to use it after the function returns (pass by reference).
        callpath_to_node_biggest = {}
        biggest_gf = biggest_gf.groupby_callpath(callpath_to_node_biggest)
        # for each callpath (i.e. node), looks at the other graphframes
        # and finds possible indices (i.e. nodes).
        for callpath, node_dict in callpath_to_node_biggest.items():
            node_in_biggest = node_dict[0]["node"]
            for gf in gf_to_visited_node.keys():
                # check if there are indices that have the
                # same name value (possible indices).
                possible_indices = gf.dataframe.loc[
                    gf.dataframe["name"]
                    == biggest_gf.dataframe.loc[node_in_biggest]["name"]
                ].index
                # iterate over possible indices (nodes).
                for node in possible_indices:
                    # get all the paths of each possible node.
                    possible_paths = node.paths()
                    num_of_paths = len(possible_paths)
                    found_count = 0
                    # for each possible path of a node
                    for possible_path in possible_paths:
                        possible_callpath = node.convert_path_to_str(possible_path)
                        # check if the possible callpath is the same with the
                        # callpath of the node in the biggest graphframe.
                        if callpath == possible_callpath:
                            found_count += 1
                    # if we visited all the paths of a node and if they are
                    # the same with the callpaths of the node in the
                    # biggest graphframe, add node to the visited list to
                    # be removed later.
                    if found_count == num_of_paths:
                        gf_to_visited_node[gf].append(node)
                        # there might be multiple nodes that have the same
                        # callpath in other graphframes as well. if we already
                        # stored any of the inc_metrics of the other graphframe
                        # before, that means we have already seen the same callpath.
                        if gf.inc_metrics[0] in node_dict:
                            # aggregate the metric values.
                            for metric in gf.inc_metrics + gf.exc_metrics:
                                node_dict[metric] += gf.dataframe.loc[
                                    node_dict["node"], metric
                                ]
                        # if not, update the node dicts (i.e. metric values) by
                        # including the metrics from the other graphframes.
                        else:
                            tmp_dict = gf.dataframe.loc[node].to_dict()
                            tmp_dict["node"] = node_in_biggest
                            callpath_to_node_biggest[callpath][0].update(tmp_dict)

        # iterate over the remaining nodes in other graphframes.
        # the remaining nodes should be created since the biggest
        # graphframe doesn't have them.
        for graphframe, visited_nodes in gf_to_visited_node.items():
            biggest_gf.inc_metrics.extend(graphframe.inc_metrics)
            biggest_gf.exc_metrics.extend(graphframe.exc_metrics)
            # for each node that is not visited in the biggest
            # graphframe.
            for node in graphframe.dataframe.index:
                if node not in visited_nodes:
                    # for each path of the node.
                    for path in node.paths():
                        # convert the callpath from list of node objects
                        # to the list of strings.
                        callpath = node.convert_path_to_str(path)
                        # the other graphframes might contain the same
                        # callpath, so we still need to check if the
                        # callpath was seen before.
                        if callpath not in callpath_to_node_biggest.keys():
                            # if not seen before, create the node and
                            # its parents if needed.
                            _create_node(
                                biggest_gf,
                                graphframe,
                                node,
                                path,
                                callpath_to_node_biggest,
                            )
                        # if already seen, update the node dicts
                        # (i.e. metric values) by including the
                        # metrics from the other graphframes.
                        else:
                            tmp_dict = graphframe.dataframe.loc[node].to_dict()
                            callpath_to_node_biggest[callpath][0].update(tmp_dict)
                    gf_to_visited_node[graphframe].append(node)

        # create a new dataframe.
        all_info = []
        for val in callpath_to_node_biggest.values():
            all_info.append(val[0])
        dataframe = pd.DataFrame(all_info)
        dataframe.set_index("node", inplace=True)
        dataframe.sort_index(inplace=True)

        # create a new graph.
        roots = biggest_gf.graph.roots
        graph = Graph(roots)
        graph.enumerate_traverse()

        for graphframe in graphframes:
            drop_columns = set(dataframe.columns) - set(graphframe.dataframe.columns)
            graphframe.dataframe = dataframe.copy()
            graphframe.dataframe.drop(columns=drop_columns, axis=1, inplace=True)
            _rename_colums(
                graphframe,
                graphframe.inc_metrics,
                graphframe.metadata[num_procs],
                add=False,
            )
            _rename_colums(
                graphframe,
                graphframe.exc_metrics,
                graphframe.metadata[num_procs],
                add=False,
            )
            graphframe.graph = graph

    def unify(self, other):
        """Returns a unified graphframe.

        Ensure self and other have the same graph and same node IDs. This may
        change the node IDs in the dataframe.

        Update the graphs in the graphframe if they differ.
        """
        if self.graph is other.graph:
            return

        node_map = {}
        union_graph = self.graph.union(other.graph, node_map)

        self_index_names = self.dataframe.index.names
        other_index_names = other.dataframe.index.names

        self.dataframe.reset_index(inplace=True)
        other.dataframe.reset_index(inplace=True)

        self.dataframe["node"] = self.dataframe["node"].apply(lambda x: node_map[id(x)])
        other.dataframe["node"] = other.dataframe["node"].apply(
            lambda x: node_map[id(x)]
        )

        # add missing rows to copy of self's dataframe in preparation for
        # operation
        self._insert_missing_rows(other)

        self.dataframe.set_index(self_index_names, inplace=True, drop=True)
        other.dataframe.set_index(other_index_names, inplace=True, drop=True)

        self.graph = union_graph
        other.graph = union_graph

    @deprecated_params(
        metric="metric_column",
        name="name_column",
        expand_names="expand_name",
        context="context_column",
        invert_colors="invert_colormap",
    )
    @Logger.loggable
    def tree(
        self,
        metric_column=None,
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    ):
        """Format this graphframe as a tree and return the resulting string."""
        color = sys.stdout.isatty()
        shell = None
        if metric_column is None:
            metric_column = self.default_metric

        if color is False:
            try:
                import IPython

                shell = IPython.get_ipython().__class__.__name__
            except ImportError:
                pass
            # Test if running in a Jupyter notebook or qtconsole
            if shell == "ZMQInteractiveShell":
                color = True

        if sys.version_info.major == 2:
            unicode = False
        elif sys.version_info.major == 3:
            unicode = True

        return ConsoleRenderer(unicode=unicode, color=color).render(
            self.graph.roots,
            self.dataframe,
            metric_column=metric_column,
            precision=precision,
            name_column=name_column,
            expand_name=expand_name,
            context_column=context_column,
            rank=rank,
            thread=thread,
            depth=depth,
            highlight_name=highlight_name,
            colormap=colormap,
            invert_colormap=invert_colormap,
        )

    @Logger.loggable
    def to_dot(self, metric=None, name="name", rank=0, thread=0, threshold=0.0):
        """Write the graph in the graphviz dot format:
        https://www.graphviz.org/doc/info/lang.html
        """
        if metric is None:
            metric = self.default_metric
        return trees_to_dot(
            self.graph.roots, self.dataframe, metric, name, rank, thread, threshold
        )

    @Logger.loggable
    def to_flamegraph(self, metric=None, name="name", rank=0, thread=0, threshold=0.0):
        """Write the graph in the folded stack output required by FlameGraph
        http://www.brendangregg.com/flamegraphs.html
        """
        folded_stack = ""
        if metric is None:
            metric = self.default_metric

        for root in self.graph.roots:
            for hnode in root.traverse():
                callpath = hnode.path()
                for i in range(0, len(callpath) - 1):
                    if (
                        "rank" in self.dataframe.index.names
                        and "thread" in self.dataframe.index.names
                    ):
                        df_index = (callpath[i], rank, thread)
                    elif "rank" in self.dataframe.index.names:
                        df_index = (callpath[i], rank)
                    elif "thread" in self.dataframe.index.names:
                        df_index = (callpath[i], thread)
                    else:
                        df_index = callpath[i]
                    folded_stack = (
                        folded_stack + str(self.dataframe.loc[df_index, "name"]) + "; "
                    )

                if (
                    "rank" in self.dataframe.index.names
                    and "thread" in self.dataframe.index.names
                ):
                    df_index = (callpath[-1], rank, thread)
                elif "rank" in self.dataframe.index.names:
                    df_index = (callpath[-1], rank)
                elif "thread" in self.dataframe.index.names:
                    df_index = (callpath[-1], thread)
                else:
                    df_index = callpath[-1]
                folded_stack = (
                    folded_stack + str(self.dataframe.loc[df_index, "name"]) + " "
                )

                # set dataframe index based on if rank and thread are part of the index
                if (
                    "rank" in self.dataframe.index.names
                    and "thread" in self.dataframe.index.names
                ):
                    df_index = (hnode, rank, thread)
                elif "rank" in self.dataframe.index.names:
                    df_index = (hnode, rank)
                elif "thread" in self.dataframe.index.names:
                    df_index = (hnode, thread)
                else:
                    df_index = hnode

                folded_stack = (
                    folded_stack + str(self.dataframe.loc[df_index, metric]) + "\n"
                )

        return folded_stack

    @Logger.loggable
    def to_literal(self, name="name", rank=0, thread=0, cat_columns=[]):
        """Format this graph as a list of dictionaries for Roundtrip
        visualizations.
        """
        graph_literal = []
        visited = []

        def _get_df_index(hnode):
            if (
                "rank" in self.dataframe.index.names
                and "thread" in self.dataframe.index.names
            ):
                df_index = (hnode, rank, thread)
            elif "rank" in self.dataframe.index.names:
                df_index = (hnode, rank)
            elif "thread" in self.dataframe.index.names:
                df_index = (hnode, thread)
            else:
                df_index = hnode

            return df_index

        def metrics_to_dict(df_index):
            metrics_dict = {}
            for m in sorted(self.inc_metrics + self.exc_metrics):
                node_metric_val = self.dataframe.loc[df_index, m]
                metrics_dict[m] = node_metric_val

            return metrics_dict

        def attributes_to_dict(df_index):
            valid_columns = [
                col for col in cat_columns if col in self.dataframe.columns
            ]

            attributes_dict = {}
            for m in sorted(valid_columns):
                node_attr_val = self.dataframe.loc[df_index, m]
                attributes_dict[m] = node_attr_val

            return attributes_dict

        def add_nodes(hnode):
            df_index = _get_df_index(hnode)

            node_dict = {}

            node_name = self.dataframe.loc[df_index, name]

            node_dict["name"] = node_name
            node_dict["frame"] = hnode.frame.attrs
            node_dict["metrics"] = metrics_to_dict(df_index)
            node_dict["metrics"]["_hatchet_nid"] = hnode._hatchet_nid
            node_dict["attributes"] = attributes_to_dict(df_index)

            if hnode.children and hnode not in visited:
                visited.append(hnode)
                node_dict["children"] = []

                for child in sorted(hnode.children, key=lambda n: n.frame):
                    node_dict["children"].append(add_nodes(child))

            return node_dict

        for root in sorted(self.graph.roots, key=lambda n: n.frame):
            graph_literal.append(add_nodes(root))

        return graph_literal

    def _operator(self, other, op):
        """Generic function to apply operator to two dataframes and store
        result in self.

        Arguments:
            self (graphframe): self's graphframe
            other (graphframe): other's graphframe
            op (operator): pandas arithmetic operator

        Return:
            (GraphFrame): self's graphframe modified
        """
        # unioned set of self and other exclusive and inclusive metrics
        all_metrics = list(
            set().union(
                self.exc_metrics, self.inc_metrics, other.exc_metrics, other.inc_metrics
            )
        )

        self.dataframe.update(op(other.dataframe[all_metrics]))

        return self

    def _insert_missing_rows(self, other):
        """Helper function to add rows that exist in other, but not in self.

        This returns a graphframe with a modified dataframe. The new rows will
        contain zeros for numeric columns.

        Return:
            (GraphFrame): self's modified graphframe
        """
        all_metrics = list(
            set().union(
                self.exc_metrics, self.inc_metrics, other.exc_metrics, other.inc_metrics
            )
        )

        # make two 2D nparrays arrays with two columns:
        # 1) the hashed value of a node and 2) a numerical index
        # Many operations are stacked here to reduce the need for storing
        # large intermediary datasets
        self_hsh_ndx = np.vstack(
            (
                np.array(
                    [x.__hash__() for x in self.dataframe["node"]], dtype=np.uint64
                ),
                self.dataframe.index.values.astype(np.uint64),
            )
        ).T
        other_hsh_ndx = np.vstack(
            (
                np.array(
                    [x.__hash__() for x in other.dataframe["node"]], dtype=np.uint64
                ),
                other.dataframe.index.values.astype(np.uint64),
            )
        ).T

        # sort our 2D arrays by hashed node value so a binary search can be used
        # in the cython function fast_not_isin
        self_hsh_ndx_sorted = self_hsh_ndx[self_hsh_ndx[:, 0].argsort()]
        other_hsh_ndx_sorted = other_hsh_ndx[other_hsh_ndx[:, 0].argsort()]

        # get nodes that exist in other, but not in self, set metric columns to 0 for
        # these rows
        other_not_in_self = other.dataframe[
            _gfm_cy.fast_not_isin(
                other_hsh_ndx_sorted,
                self_hsh_ndx_sorted,
                other_hsh_ndx_sorted.shape[0],
                self_hsh_ndx_sorted.shape[0],
            )
        ]
        # get nodes that exist in self, but not in other
        self_not_in_other = self.dataframe[
            _gfm_cy.fast_not_isin(
                self_hsh_ndx_sorted,
                other_hsh_ndx_sorted,
                self_hsh_ndx_sorted.shape[0],
                other_hsh_ndx_sorted.shape[0],
            )
        ]

        # if there are missing nodes in either self or other, add a new column
        # called _missing_node
        if not self_not_in_other.empty:
            self.dataframe = self.dataframe.assign(
                _missing_node=np.zeros(len(self.dataframe), dtype=np.short)
            )
        if not other_not_in_self.empty:
            # initialize with 2 to save filling in later
            other_not_in_self = other_not_in_self.assign(
                _missing_node=[int(2) for x in range(len(other_not_in_self))]
            )

            # add a new column to self if other has nodes not in self
            if self_not_in_other.empty:
                self.dataframe["_missing_node"] = np.zeros(
                    len(self.dataframe), dtype=np.short
                )

        # get lengths to pass into
        onis_len = len(other_not_in_self)
        snio_len = len(self_not_in_other)

        # case where self is a superset of other
        if snio_len != 0:
            self_missing_node = self.dataframe["_missing_node"].values
            snio_indices = self_not_in_other.index.values

            # This function adds 1 to all nodes in self.dataframe['_missing_node'] which
            # are in self but not in the other graphframe
            _gfm_cy.insert_one_for_self_nodes(snio_len, self_missing_node, snio_indices)
            self.dataframe["_missing_node"] = np.array(
                [n for n in self_missing_node], dtype=np.short
            )

        # for nodes that only exist in other, set the metric to be nan (since
        # it's a missing node in self)
        # replaces individual metric assignments with np.zeros
        for j in all_metrics:
            other_not_in_self[j] = np.full(onis_len, np.nan)

        # append missing rows (nodes that exist in other, but not in self) to self's
        # dataframe
        self.dataframe = pd.concat(
            [self.dataframe, other_not_in_self], axis=0, sort=True
        )

        return self

    @Logger.loggable
    def groupby_aggregate(self, groupby_column, agg_function):
        """Groupby-aggregate dataframe and reindex the Graph.

        Reindex the graph to match the groupby-aggregated dataframe.

        Update the frame attributes to contain those columns in the dataframe index.

        Arguments:
            self (graphframe): self's graphframe
            groupby_column: column to groupby on dataframe
            agg_function: aggregate function on dataframe

        Return:
            (GraphFrame): new graphframe with reindexed graph and groupby-aggregated dataframe
        """
        # create new nodes for each unique node in the old dataframe
        # length is equal to number of nodes in original graph
        old_to_new = {}

        # list of new roots
        new_roots = []

        # dict of (new) super nodes
        # length is equal to length of dataframe index (after groupby-aggregate)
        node_dicts = []

        def reindex(node, parent, visited):
            """Reindex the graph.

            Connect super nodes to children according to relationships from old graph.
            """
            # grab the super node corresponding to original node
            super_node = old_to_new.get(node)

            if not node.parents and super_node not in new_roots:
                # this is a new root
                new_roots.append(super_node)

            # iterate over parents of old node, adding parents to super node
            for parent in node.parents:
                # convert node to super node
                snode = old_to_new.get(parent)
                # move to next node if parent and super node are to be merged
                if snode == super_node:
                    continue
                # add node to super node's parents if parent does not exist in super
                # node's parents
                if snode not in super_node.parents:
                    super_node.add_parent(snode)

            # iterate over children of old node, adding children to super node
            for child in node.children:
                # convert node to super node
                snode = old_to_new.get(child)
                # move to next node if child and super node are to be merged
                if snode == super_node:
                    continue
                # add node to super node's children if child does not exist in super
                # node's children
                if snode not in super_node.children:
                    super_node.add_child(snode)

            if node not in visited:
                visited.add(node)
                for child in node.children:
                    reindex(child, super_node, visited)

        # groupby-aggregate dataframe based on user-supplied functions
        groupby_obj = self.dataframe.groupby(groupby_column)
        agg_df = groupby_obj.agg(agg_function)

        # traverse groupby_obj, determine old node to super node mapping
        nid = 0
        for k, v in groupby_obj.groups.items():
            node_name = k
            node_type = agg_df.index.name
            super_node = Node(Frame({"name": node_name, "type": node_type}), None, nid)
            n = {"node": super_node, "nid": nid, "name": node_name}
            node_dicts.append(n)
            nid += 1

            # if many old nodes map to the same super node
            for i in v:
                old_to_new[i] = super_node

        # reindex graph by traversing old graph
        visited = set()
        for root in self.graph.roots:
            reindex(root, None, visited)

        # append super nodes to groupby-aggregate dataframe
        df_index = list(agg_df.index.names)
        agg_df.reset_index(inplace=True)
        df_nodes = pd.DataFrame.from_dict(data=node_dicts)
        tmp_df = pd.concat([agg_df, df_nodes], axis=1)
        # add node to dataframe index if it doesn't exist
        if "node" not in df_index:
            df_index.append("node")
        # reset index
        tmp_df.set_index(df_index, inplace=True)

        # update _hatchet_nid in reindexed graph and groupby-aggregate dataframe
        graph = Graph(new_roots)
        graph.enumerate_traverse()

        # put it all together
        new_gf = GraphFrame(
            graph,
            tmp_df,
            list(self.exc_metrics),
            list(self.inc_metrics),
            self.default_metric,
            dict(self.metadata),
            attributes=dict([[x, getattr(self, x)] for x in self.attributes]),
        )
        new_gf.drop_index_levels()
        return new_gf

    @Logger.loggable
    def flat_profile(self, groupby_column=None, as_index=True):
        """Generates flat profile for a given graphframe.
        Returns a new dataframe."""
        return Chopper().flat_profile(self, groupby_column, as_index)

    @Logger.loggable
    def flatten(self, groupby_column=None):
        """
        Flattens the graphframe by changing its graph structure and the dataframe.
        """
        return Chopper().flatten(self, groupby_column)

    @Logger.loggable
    def to_callgraph(self):
        """
        Converts a CCT to a callgraph.
        Returns a new graphframe.
        """
        return Chopper().to_callgraph(self)

    @Logger.loggable
    def load_imbalance(self, metric_column=None, threshold=None, verbose=False):
        """Calculates load imbalance for given metric column(s)
        Takes a graphframe and a list of metric column(s), and
        returns a new graphframe with metric.imbalance column(s).
        """
        return Chopper().load_imbalance(self, metric_column, threshold, verbose)

    @Logger.loggable
    def hot_path(self, start_node=None, metric=None, threshold=0.5):
        """Returns the hot_path function.
        Inputs:
         - start_node: Start node of the hot path should be given.
         - metric: A numerical metric on the dataframe
         - threshold: Threshold for parent-child comparison (parent <= child/2).
        Output:
         - hot_path: list of nodes, starting from the start node to the hot node.

        Example:
        root_node = graphframe.graph.roots[0]
        graphframe.hot_path(root_node)
        """
        # call hot_path function on high-level API
        hot_path = Chopper().hot_path(
            self, start_node, metric, threshold, callpath=[start_node]
        )
        return hot_path

    @Logger.loggable
    def correlation_analysis(self, metrics=None, method="spearman"):
        """Calculates correlation between metrics of a given graphframe.
        Returns the correlation matrix.

        Pandas provides three different methods: pearson, spearman, kendall
        """
        # call correlation_analysis function on high-level API
        correlation_matrix = Chopper().correlation_analysis(self, metrics, method)
        return correlation_matrix

    @Logger.loggable
    def add(self, other):
        """Returns the column-wise sum of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.add)

    @Logger.loggable
    def sub(self, other):
        """Returns the column-wise difference of two graphframes as a new
        graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.sub)

    @Logger.loggable
    def div(self, other):
        """Returns the column-wise float division of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.divide)

    @Logger.loggable
    def mul(self, other):
        """Returns the column-wise float multiplication of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        # create a copy of both graphframes
        self_copy = self.copy()
        other_copy = other.copy()

        # unify copies of graphframes
        self_copy.unify(other_copy)

        return self_copy._operator(other_copy, self_copy.dataframe.multiply)

    def __iadd__(self, other):
        """Computes column-wise sum of two graphframes and stores the result in
        self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and copy of other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.add)

    def __add__(self, other):
        """Returns the column-wise sum of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.add(other)

    def __mul__(self, other):
        """Returns the column-wise multiplication of two graphframes as a new graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.mul(other)

    def __isub__(self, other):
        """Computes column-wise difference of two graphframes and stores the
        result in self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.sub)

    def __sub__(self, other):
        """Returns the column-wise difference of two graphframes as a new
        graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.sub(other)

    def __idiv__(self, other):
        """Computes column-wise float division of two graphframes and stores the
        result in self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.div)

    def __truediv__(self, other):
        """Returns the column-wise float division of two graphframes as a new
        graphframe.

        This graphframe is the union of self's and other's graphs, and does not
        modify self or other.

        Return:
            (GraphFrame): new graphframe
        """
        return self.div(other)

    def __imul__(self, other):
        """Computes column-wise float multiplication of two graphframes and stores the
        result in self.

        Self's graphframe is the union of self's and other's graphs, and the
        node handles from self will be rewritten with this operation. This
        operation does not modify other.

        Return:
            (GraphFrame): self's graphframe modified
        """
        # create a copy of other's graphframe
        other_copy = other.copy()

        # unify self graphframe and other graphframe
        self.unify(other_copy)

        return self._operator(other_copy, self.dataframe.mul)


class InvalidFilter(Exception):
    """Raised when an invalid argument is passed to the filter function."""


class EmptyFilter(Exception):
    """Raised when a filter would otherwise return an empty GraphFrame."""



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

dot_keywords = ["graph", "subgraph", "digraph", "node", "edge", "strict"]



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import subprocess
import numpy as np

import pytest
import sys

from hatchet import GraphFrame
from hatchet.readers.caliper_reader import CaliperReader
from hatchet.util.executable import which
from hatchet.external.console import ConsoleRenderer

caliperreader_avail = True
try:
    import caliperreader
except ImportError:
    caliperreader_avail = False

annotations = [
    "main",
    "LagrangeLeapFrog",
    "LagrangeElements",
    "ApplyMaterialPropertiesForElems",
    "EvalEOSForElems",
    "CalcEnergyForElems",
    "CalcPressureForElems",
    "CalcSoundSpeedForElems",
    "UpdateVolumesForElems",
    "CalcTimeConstraintsForElems",
    "CalcCourantConstraintForElems",
    "CalcHydroConstraintForElems",
    "TimeIncrement",
    "LagrangeNodal",
    "CalcForceForNodes",
    "CalcVolumeForceForElems",
    "IntegrateStressForElems",
    "CalcHourglassControlForElems",
    "CalcFBHourglassForceForElems",
    "CalcLagrangeElements",
    "CalcKinematicsForElems",
    "CalcQForElems",
    "CalcMonotonicQGradientsForElems",
    "CalcMonotonicQRegionForElems",
]


def test_graphframe(lulesh_caliper_json):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_caliper(str(lulesh_caliper_json))

    assert len(gf.dataframe.groupby("name")) == 24

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("nid", "rank"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "node"):
            assert gf.dataframe[col].dtype == object

    # TODO: add tests to confirm values in dataframe


def test_read_lulesh_json(lulesh_caliper_json):
    """Sanity check the Caliper reader by examining a known input."""
    reader = CaliperReader(str(lulesh_caliper_json))
    reader.read_json_sections()

    assert len(reader.json_data) == 192
    assert len(reader.json_cols) == 4
    assert len(reader.json_cols_mdata) == 4
    assert len(reader.json_nodes) == 24

    reader.create_graph()
    assert all(an in reader.idx_to_label.values() for an in annotations)


def test_calc_pi_json(calc_pi_caliper_json):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_caliper(str(calc_pi_caliper_json))

    assert len(gf.dataframe.groupby("name")) == 100


@pytest.mark.skipif(not which("cali-query"), reason="needs cali-query to be in path")
def test_lulesh_cali(lulesh_caliper_cali):
    """Sanity check the Caliper reader ingesting a .cali file."""
    grouping_attribute = "function"
    default_metric = "sum(sum#time.duration),inclusive_sum(sum#time.duration)"
    query = "select function,%s group by %s format json-split" % (
        default_metric,
        grouping_attribute,
    )

    gf = GraphFrame.from_caliper(str(lulesh_caliper_cali), query)

    assert len(gf.dataframe.groupby("name")) == 18


@pytest.mark.skipif(not which("cali-query"), reason="needs cali-query to be in path")
def test_lulesh_json_stream(lulesh_caliper_cali):
    """Sanity check the Caliper reader ingesting a JSON string literal."""
    cali_query = which("cali-query")
    grouping_attribute = "function"
    default_metric = "sum(sum#time.duration),inclusive_sum(sum#time.duration)"
    query = "select function,%s group by %s format json-split" % (
        default_metric,
        grouping_attribute,
    )

    cali_json = subprocess.Popen(
        [cali_query, "-q", query, lulesh_caliper_cali], stdout=subprocess.PIPE
    )

    gf = GraphFrame.from_caliper(cali_json.stdout)

    assert len(gf.dataframe.groupby("name")) == 18


@pytest.mark.skipif(sys.version_info > (3, 8), reason="Temporarily let this test fail.")
def test_filter_squash_unify_caliper_data(lulesh_caliper_json):
    """Sanity test a GraphFrame object with known data."""
    gf1 = GraphFrame.from_caliper(str(lulesh_caliper_json))
    gf2 = GraphFrame.from_caliper(str(lulesh_caliper_json))

    assert gf1.graph is not gf2.graph

    gf1_index_names = gf1.dataframe.index.names
    gf2_index_names = gf2.dataframe.index.names

    gf1.dataframe.reset_index(inplace=True)
    gf2.dataframe.reset_index(inplace=True)

    # indexes are the same since we are reading in the same dataset
    assert all(gf1.dataframe["node"] == gf2.dataframe["node"])

    gf1.dataframe.set_index(gf1_index_names, inplace=True)
    gf2.dataframe.set_index(gf2_index_names, inplace=True)

    squash_gf1 = gf1.filter(lambda x: x["name"].startswith("Calc"))
    squash_gf2 = gf2.filter(lambda x: x["name"].startswith("Calc"))

    squash_gf1.unify(squash_gf2)

    assert squash_gf1.graph is squash_gf2.graph

    squash_gf1.dataframe.reset_index(inplace=True)
    squash_gf2.dataframe.reset_index(inplace=True)

    # Indexes should still be the same after unify. Sort indexes before comparing.
    assert all(squash_gf1.dataframe["node"] == squash_gf2.dataframe["node"])

    squash_gf1.dataframe.set_index(gf1_index_names, inplace=True)
    squash_gf2.dataframe.set_index(gf2_index_names, inplace=True)


def test_tree(lulesh_caliper_json):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_caliper(str(lulesh_caliper_json))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "121489.000 main" in output
    assert "663.000 LagrangeElements" in output
    assert "21493.000 CalcTimeConstraintsForElems" in output

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time (inc)",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "662712.000 EvalEOSForElems" in output
    assert "2895319.000 LagrangeNodal" in output


def test_graphframe_to_literal(lulesh_caliper_json):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_caliper(str(lulesh_caliper_json))
    graph_literal = gf.to_literal()

    gf2 = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf2.graph)


def test_graphframe_native_lulesh_from_file(lulesh_caliper_cali):
    """Sanity check the native Caliper reader by examining a known input."""

    gf = GraphFrame.from_caliperreader(str(lulesh_caliper_cali))

    assert len(gf.dataframe.groupby("name")) == 19
    assert "cali.caliper.version" in gf.metadata.keys()

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("nid", "rank"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "node"):
            assert gf.dataframe[col].dtype == object


@pytest.mark.skipif(
    not caliperreader_avail, reason="needs caliper-reader package to be loaded"
)
def test_graphframe_native_lulesh_from_caliperreader(lulesh_caliper_cali):
    """Sanity check the native Caliper reader by examining a known input."""
    r = caliperreader.CaliperReader()
    r.read(lulesh_caliper_cali)

    gf = GraphFrame.from_caliperreader(r)

    assert len(gf.dataframe.groupby("name")) == 19
    assert "cali.caliper.version" in gf.metadata.keys()

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("nid", "rank"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "node"):
            assert gf.dataframe[col].dtype == object


def test_inclusive_time_calculation(lulesh_caliper_json):
    """Validate calculate_inclusive_metrics() on known dataset containing per-rank data."""
    gf = GraphFrame.from_caliper(str(lulesh_caliper_json))

    # save original time (inc) column for correctness check
    gf.dataframe["orig_inc_time"] = gf.dataframe["time (inc)"]

    # remove original time (inc) column since it will be generated by calculate_inclusive_metrics()
    del gf.dataframe["time (inc)"]

    gf.calculate_inclusive_metrics()
    assert all(
        gf.dataframe["time (inc)"].values == gf.dataframe["orig_inc_time"].values
    )


def test_sw4_cuda_from_caliperreader(sw4_caliper_cuda_activity_profile_cali):
    gf = GraphFrame.from_caliperreader(sw4_caliper_cuda_activity_profile_cali)

    assert len(gf.graph) == 549
    assert all(
        metric in gf.dataframe.columns for metric in gf.exc_metrics + gf.inc_metrics
    )

    for col in gf.dataframe.columns:
        if col in ("#scale#cupti.activity.duration", "#scale#sum#cupti.host.duration"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in "rank":
            assert gf.dataframe[col].dtype == np.int64
        elif col in "name":
            assert gf.dataframe[col].dtype == object

    for col in gf.exc_metrics + gf.inc_metrics:
        assert col in gf.dataframe.columns


def test_sw4_cuda_summary_from_caliperreader(
    sw4_caliper_cuda_activity_profile_summary_cali,
):
    gf = GraphFrame.from_caliperreader(sw4_caliper_cuda_activity_profile_summary_cali)

    assert len(gf.graph) == 393
    assert all(
        metric in gf.dataframe.columns for metric in gf.exc_metrics + gf.inc_metrics
    )

    for col in gf.exc_metrics + gf.inc_metrics:
        assert col in gf.dataframe.columns



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd
import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame

# Requires pycubexr >= v1.2.0
from pycubexr.parsers.tar_parser import CubexParser
from pycubexr.utils.exceptions import MissingMetricError


class ScorePReader:
    """Read in cubex files generated using Scorep."""

    def __init__(self, filename):
        self.filename = filename
        self.callpath_to_node_dicts = {}
        self.callpath_to_node = {}
        self.inc_metrics = []
        self.exc_metrics = []
        self.list_roots = []
        self.threads_per_rank = 0
        self.multiple_ranks = False
        self.multiple_threads = False

    def create_graph(self, cubex):
        def _create_node_dict(
            node_dict,
            callpath_rank_thread,
            location,
            metric,
            metric_values,
            hatchet_node,
            pycubexr_cnode,
            name,
            begin_line,
            end_line,
            file,
        ):
            """Gets all the related information about a node
            and creates a node dictionary to be added to
            the dataframe"""
            if callpath_rank_thread not in node_dict:
                node_dict[callpath_rank_thread] = {
                    "node": hatchet_node,
                    "name": name,
                    "rank": callpath_rank_thread[1],
                    "thread": callpath_rank_thread[2],
                    "line": begin_line,
                    "end_line": end_line,
                    "file": file,
                }

            # example metrics: 'name', 'visits', 'time', 'min_time',
            # 'max_time', 'bytes_sent', 'bytes_received'.
            metric_name = metric.name
            metric_value = None
            # MinValue and MaxValue require special consideration because pycubexr
            # stores them as objects instead of floats. We convert them to float.
            # We store the value as it is if it is not MinValue or MaxValue.
            if metric_name == "min_time" or metric_name == "max_time":
                metric_value = metric_values.location_value(
                    pycubexr_cnode, location.id
                ).value
            else:
                metric_value = metric_values.location_value(pycubexr_cnode, location.id)

            # We should check each metric and add ' (inc)' if
            # their type is 'INCLUSIVE'.
            if metric.metric_type == "INCLUSIVE":
                metric_name = "{}{}".format(metric.name, " (inc)")

            node_dict[callpath_rank_thread].update({metric_name: metric_value})

        def _calculate_rank_thread(location):
            """Calculates and returns the rank and thread number using
            the location information stored in Score-P. Location ID is
            a number between 0 and (# of ranks) * (# of threads) - 1.
            Also, checks if multiple ranks or threads exists to be
            able to create the dataframe accordingly."""
            # calculate rank and thread numbers.
            # we should do this calculation for each location.id.
            # location.id is a number from 0 to (# of ranks) * (# of threads) - 1
            rank = location.id // self.threads_per_rank
            # thread number is stored as rank in pyCubexR
            thread = int(location.rank)
            if not self.multiple_ranks and rank != 0:
                self.multiple_ranks = True
            if not self.multiple_threads and thread != 0:
                self.multiple_threads = True
            return (rank, thread)

        def _get_node_info(pycubexr_cnode, node_name, begin_line, end_line, file):
            """Gets the node name, begin and end line, and file info
            of a node from pycubexr."""
            node_name = cubex.get_region(pycubexr_cnode).name
            begin_line = cubex.get_region(pycubexr_cnode).begin
            end_line = cubex.get_region(pycubexr_cnode).end
            file = cubex.get_region(pycubexr_cnode).mod
            return node_name, begin_line, end_line, file

        def _get_callpath_info(pycubexr_cnode, parent_callpath):
            """Gets the callpath information by traversing the
            tree created by pycubexr.
            Checks if callpath of a node is already existing.
            If yes, does not create a new node. Otherwise, creates
            a new node and sets parent-child relationships."""
            node_dict = {}
            parent_node = self.callpath_to_node[parent_callpath]
            node_name, begin_line, end_line, file = None, None, None, None
            node_name, begin_line, end_line, file = _get_node_info(
                pycubexr_cnode, node_name, begin_line, end_line, file
            )
            callpath = parent_callpath + (node_name,)
            callpath_rank_thread = tuple()

            if callpath not in self.callpath_to_node:
                # Create the root node since it doesn't exist
                hatchet_node = Node(
                    Frame({"name": node_name, "type": "function"}), None
                )

                # Store callpaths to identify nodes
                self.callpath_to_node[callpath] = hatchet_node
                hatchet_node.add_parent(parent_node)
                parent_node.add_child(hatchet_node)
            else:
                # Don't create a new node since it is created earlier
                hatchet_node = self.callpath_to_node.get(callpath)

            # this for loop and try-except statements are taken from
            # pycubexr readme file on github.
            for metric in cubex.get_metrics():
                try:
                    # get metric values for each metric.
                    metric_values = cubex.get_metric_values(metric)
                    for location in cubex.get_locations():
                        # get (callpath, rank, thread) tuple for each rank or thread
                        callpath_rank_thread = (callpath,) + _calculate_rank_thread(
                            location
                        )

                        _create_node_dict(
                            node_dict,
                            callpath_rank_thread,
                            location,
                            metric,
                            metric_values,
                            hatchet_node,
                            pycubexr_cnode,
                            node_name,
                            begin_line,
                            end_line,
                            file,
                        )
                except MissingMetricError:
                    # Ignore missing metrics on Score-P
                    pass

            # Sometimes pyCubexR stores some nodes with the same
            # callpath as different nodes even though the only
            # different is their metric values. Here we check if
            # we have already seen the (callpath, rank, thread) and
            # aggregate their values if we have seen it instead of
            # creating a new node not to get ambigous data error
            # on the dataframe.
            if callpath_rank_thread in self.callpath_to_node_dicts:
                for c_r_t in node_dict.keys():
                    for key, value in node_dict[c_r_t].items():
                        if key in self.inc_metrics or key in self.exc_metrics:
                            self.callpath_to_node_dicts[c_r_t][key] += value
            else:
                self.callpath_to_node_dicts.update(node_dict)

            for child in pycubexr_cnode.get_children():
                _get_callpath_info(child, callpath)

        root = cubex._anchor_result.cnodes[0]
        root_name, root_begin, root_end, root_file = None, None, None, None
        root_name, root_begin, root_end, root_file = _get_node_info(
            root, root_name, root_begin, root_end, root_file
        )

        callpath = (root_name,)
        if callpath not in self.callpath_to_node:
            # Create the root node since it doesn't exist
            root_node = Node(Frame({"name": root_name, "type": "function"}), None)

            # Store callpaths to identify nodes
            self.callpath_to_node[callpath] = root_node
            self.list_roots.append(root_node)
        else:
            # Don't create a new node since it is created earlier
            root_node = self.callpath_to_node.get(callpath)

        node_dict = {}
        # get_locations() gets all the rank/thread information.
        # last location refers to the last thread information.
        # if there are 8 threads, the last one should return 7.
        self.threads_per_rank = int(cubex.get_locations()[-1].rank) + 1
        for metric in cubex.get_metrics():
            try:
                # example metrics: 'name', 'visits', 'time (inc)', 'min_time',
                # 'max_time', 'bytes_sent', 'bytes_received'.
                metric_values = cubex.get_metric_values(metric)
                if metric.metric_type == "INCLUSIVE":
                    self.inc_metrics.append(metric.name + " (inc)")
                else:
                    self.exc_metrics.append(metric.name)

                for location in cubex.get_locations():
                    callpath_rank_thread = (callpath,) + _calculate_rank_thread(
                        location
                    )
                    _create_node_dict(
                        node_dict,
                        callpath_rank_thread,
                        location,
                        metric,
                        metric_values,
                        root_node,
                        root,
                        root_name,
                        root_begin,
                        root_end,
                        root_file,
                    )
            except MissingMetricError:
                # Ignore missing metrics
                pass

        self.callpath_to_node_dicts.update(node_dict)
        for child in root.get_children():
            _get_callpath_info(child, callpath)

        return self.list_roots

    def read(self):
        cubex = CubexParser(self.filename).__enter__()
        # Add all nodes and roots.
        roots = self.create_graph(cubex)
        # Create a graph object once all nodes have been added.
        graph = Graph(roots)
        graph.enumerate_traverse()

        dataframe = pd.DataFrame.from_dict(data=self.callpath_to_node_dicts.values())

        indices = []
        # Set indices according to rank/thread numbers.
        if self.multiple_ranks and self.multiple_threads:
            indices = ["node", "rank", "thread"]
        elif self.multiple_ranks:
            dataframe.drop(columns=["thread"], inplace=True)
            indices = ["node", "rank"]
        elif self.multiple_threads:
            dataframe.drop(columns=["rank"], inplace=True)
            indices = ["node", "thread"]
        else:
            indices = ["node"]

        dataframe.set_index(indices, inplace=True)
        dataframe.sort_index(inplace=True)

        # Fill the missing ranks
        # After unstacking and iterating over rows, there
        # will be "NaN" values for some ranks. Find the first
        # rank that has notna value and use it for other rows/ranks
        # of the multiindex.
        # TODO: iterrows() is not the best way to iterate over rows.
        if self.multiple_ranks or self.multiple_threads:
            dataframe = dataframe.unstack()
            for idx, row in dataframe.iterrows():
                # There is always a valid name for an index.
                # Take that valid name and assign to other ranks/rows.
                name = row["name"][row["name"].first_valid_index()]
                dataframe.loc[idx, "name"] = name

                # Fill the rest with 0
                dataframe.fillna(0, inplace=True)

            # Stack the dataframe
            dataframe = dataframe.stack()

        # The root node (above the main function) is just the
        # name of the program and does not represent a
        # function, statement or loop.
        # In the raw score-p data, the root node has a time of 0
        # but to be consistent, we change that to the sum of
        # time values of its children.
        if "time (inc)" in self.inc_metrics:
            for root in graph.roots:
                total_val = 0
                for child in root.children:
                    total_val += dataframe.loc[child, "time (inc)"].to_numpy()
                dataframe.loc[root, "time (inc)"] = total_val

        gf = hatchet.graphframe.GraphFrame(
            graph, dataframe, self.exc_metrics, self.inc_metrics
        )

        # Exclusive time metric is missing in Score-P output.
        if "time (inc)" in self.inc_metrics:
            gf.calculate_exclusive_metrics("time (inc)")

        return gf



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet import GraphFrame


def test_graphframe(mock_graph_literal):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_literal(mock_graph_literal)

    assert len(gf.dataframe) == 24


def test_graphframe_to_literal(mock_graph_literal):
    gf = GraphFrame.from_literal(mock_graph_literal)
    graph_literal = gf.to_literal()

    test_literal_output = gf.from_literal(graph_literal)

    assert len(test_literal_output.graph.roots) == len(gf.graph.roots)
    assert len(test_literal_output.graph) == len(gf.graph)


def test_with_duplicates(mock_graph_literal_duplicates):
    gf = GraphFrame.from_literal(mock_graph_literal_duplicates)

    assert len(gf.graph) == 6

    graph_literal = gf.to_literal()
    assert mock_graph_literal_duplicates.sort() == graph_literal.sort()


def test_with_duplicate_in_first_node(mock_graph_literal_duplicate_first):
    gf = GraphFrame.from_literal(mock_graph_literal_duplicate_first)

    assert len(gf.graph) == 6

    graph_literal = gf.to_literal()
    assert mock_graph_literal_duplicate_first.sort() == graph_literal.sort()


def test_inclusive_time_calculation_mock_graph(mock_graph_literal):
    """Validate calculate_inclusive_metrics() on known dataset."""
    gf = GraphFrame.from_literal(mock_graph_literal)

    # save original time (inc) column for correctness check
    gf.dataframe["orig_inc_time"] = gf.dataframe["time (inc)"]

    # remove original time (inc) column since it will be generated by calculate_inclusive_metrics()
    del gf.dataframe["time (inc)"]

    gf.calculate_inclusive_metrics()
    assert all(
        gf.dataframe["time (inc)"].values == gf.dataframe["orig_inc_time"].values
    )


def test_inclusive_time_calculation_small_mock_graphs(
    small_mock1, small_mock2, small_mock3
):
    gf1 = GraphFrame.from_literal(small_mock1)
    gf1.dataframe["orig_inc_time"] = gf1.dataframe["time (inc)"]
    del gf1.dataframe["time (inc)"]
    gf1.calculate_inclusive_metrics()
    assert all(
        gf1.dataframe["time (inc)"].values == gf1.dataframe["orig_inc_time"].values
    )

    gf2 = GraphFrame.from_literal(small_mock2)
    gf2.dataframe["orig_inc_time"] = gf2.dataframe["time (inc)"]
    del gf2.dataframe["time (inc)"]
    gf2.calculate_inclusive_metrics()
    assert all(
        gf2.dataframe["time (inc)"].values == gf2.dataframe["orig_inc_time"].values
    )

    gf3 = GraphFrame.from_literal(small_mock3)
    gf3.dataframe["orig_inc_time"] = gf3.dataframe["time (inc)"]
    del gf3.dataframe["time (inc)"]
    gf3.calculate_inclusive_metrics()
    assert all(
        gf3.dataframe["time (inc)"].values == gf3.dataframe["orig_inc_time"].values
    )


def test_inclusive_time_calculation_mock_dag_graphs(
    mock_dag_literal1,
    mock_dag_literal2,
):
    gf4 = GraphFrame.from_literal(mock_dag_literal1)
    gf4.dataframe["orig_inc_time"] = gf4.dataframe["time (inc)"]
    del gf4.dataframe["time (inc)"]
    gf4.calculate_inclusive_metrics()
    assert all(
        gf4.dataframe["time (inc)"].values == gf4.dataframe["orig_inc_time"].values
    )

    gf5 = GraphFrame.from_literal(mock_dag_literal2)
    gf5.dataframe["orig_inc_time"] = gf5.dataframe["time (inc)"]
    del gf5.dataframe["time (inc)"]
    gf5.calculate_inclusive_metrics()
    assert all(
        gf5.dataframe["time (inc)"].values == gf5.dataframe["orig_inc_time"].values
    )


def test_inclusive_time_calculation_mock_dag_modules(
    mock_dag_literal_module,
    mock_dag_literal_module_complex,
    mock_dag_literal_module_more_complex,
):
    gf6 = GraphFrame.from_literal(mock_dag_literal_module)
    gf6.dataframe["orig_inc_time"] = gf6.dataframe["time (inc)"]
    del gf6.dataframe["time (inc)"]
    gf6.calculate_inclusive_metrics()
    assert all(
        gf6.dataframe["time (inc)"].values == gf6.dataframe["orig_inc_time"].values
    )

    gf7 = GraphFrame.from_literal(mock_dag_literal_module_complex)
    gf7.dataframe["orig_inc_time"] = gf7.dataframe["time (inc)"]
    del gf7.dataframe["time (inc)"]
    gf7.calculate_inclusive_metrics()
    assert all(
        gf7.dataframe["time (inc)"].values == gf7.dataframe["orig_inc_time"].values
    )

    gf8 = GraphFrame.from_literal(mock_dag_literal_module_more_complex)
    gf8.dataframe["orig_inc_time"] = gf8.dataframe["time (inc)"]
    del gf8.dataframe["time (inc)"]
    gf8.calculate_inclusive_metrics()
    assert all(
        gf8.dataframe["time (inc)"].values == gf8.dataframe["orig_inc_time"].values
    )



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import numpy as np

from hatchet import GraphFrame
from hatchet.external.console import ConsoleRenderer


def test_graphframe(tau_profile_dir):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_tau(str(tau_profile_dir))

    for col in gf.dataframe.columns:
        if col in ("time (inc)", "time"):
            assert gf.dataframe[col].dtype == np.float64
        elif col in ("line"):
            assert gf.dataframe[col].dtype == np.int64
        elif col in ("name", "node"):
            assert gf.dataframe[col].dtype == object

    # TODO: add tests to confirm values in dataframe


def test_tree(tau_profile_dir):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_tau(str(tau_profile_dir))

    # check the tree for rank 0
    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "449.000 .TAU application" in output
    assert "4458.000 MPI_Finalize()" in output
    assert "218.000 MPI_Bcast()" in output

    # check the tree for rank 1
    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="",
        rank=1,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )
    assert "419.000 .TAU application" in output
    assert "4894.000 MPI_Finalize()" in output
    assert "333.000 MPI_Bcast()" in output


def test_children(tau_profile_dir):
    gf = GraphFrame.from_tau(str(tau_profile_dir))
    root = gf.graph.roots
    root_children = [
        "MPI_Init()",
        "MPI_Comm_size()",
        "MPI_Comm_rank()",
        "MPI_Get_processor_name()",
        "MPI_Bcast()",
        "MPI_Reduce()",
        "MPI_Finalize()",
    ]

    # check if only one root node is created
    assert len(root) == 1

    # check if root has right children #TODO: can be improved
    for child in root[0].children:
        assert child.frame["name"] in root_children


def test_graphframe_to_literal(tau_profile_dir):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_tau(str(tau_profile_dir))
    graph_literal = gf.to_literal()

    gf_literal = GraphFrame.from_literal(graph_literal)

    assert len(gf.graph) == len(gf_literal.graph)



# Copyright 2021-2024 University of Maryland and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT


class ColorMaps:
    # RdYlGn (Default) color map
    RdYlGn = [
        "\033[38;5;196m",  # red
        "\033[38;5;208m",  # orange
        "\033[38;5;220m",  # yellow-ish
        "\033[38;5;46m",  # neon green
        "\033[38;5;34m",  # light green
        "\033[38;5;22m",  # dark green
    ]

    # BrBG color map
    BrBG = [
        "\033[38;5;94m",  # Brown
        "\033[38;5;179m",  # Light Brown
        "\033[38;5;222m",  # Cream
        "\033[38;5;116m",  # Light Turqouise
        "\033[38;5;37m",  # Blue-Green 73
        "\033[38;5;23m",  # Dark Blue-Green
    ]

    #  PiYG color map
    PiYG = [
        "\033[38;5;162m",  # Dark Pink
        "\033[38;5;176m",  # Light-ish Pink
        "\033[38;5;219m",  # Pale Pink
        "\033[38;5;149m",  # Pale Green
        "\033[38;5;70m",  # Light Forest Green
        "\033[38;5;22m",  # Dark Green
    ]

    # PRGn color map
    PRGn = [
        "\033[38;5;90m",  # Dark Purple
        "\033[38;5;140m",  # Purple
        "\033[38;5;183m",  # Light Purple
        "\033[38;5;151m",  # Pale Green
        "\033[38;5;70m",  # Forest Green
        "\033[38;5;22m",  # Dark Green
    ]

    # PuOr color map
    PuOr = [
        "\033[38;5;130m",  # Brown
        "\033[38;5;208m",  # Orange
        "\033[38;5;220m",  # Pale Orange
        "\033[38;5;189m",  # Pale Purple
        "\033[38;5;104m",  # Light-ish Purple
        "\033[38;5;57m",  # Dark Purple
    ]

    # RdBu color map
    RdBu = [
        "\033[38;5;124m",  # Maroon Red
        "\033[38;5;209m",  # Dark Orange
        "\033[38;5;224m",  # Pale Orange
        "\033[38;5;153m",  # Light Sky Blue
        "\033[38;5;75m",  # Light-ish Blue
        "\033[38;5;25m",  # Navy Blue
    ]

    # RdGy color map
    RdGy = [
        "\033[38;5;124m",  # Maroon Red
        "\033[38;5;209m",  # Dark Orange
        "\033[38;5;223m",  # Pale Orange
        "\033[38;5;251m",  # Pale Grey
        "\033[38;5;244m",  # Grey
        "\033[38;5;238m",  # Dark Grey
    ]

    # RdYlBu color map
    RdYlBu = [
        "\033[38;5;196m",  # Red
        "\033[38;5;208m",  # Orange
        "\033[38;5;220m",  # Yellow-ish
        "\033[38;5;153m",  # Light Cyan Blue
        "\033[38;5;68m",  # Seagull Blue
        "\033[38;5;24m",  # Dark Blue
    ]

    # Spectral color map
    Spectral = [
        "\033[38;5;196m",  # Red
        "\033[38;5;208m",  # Orange
        "\033[38;5;220m",  # Yellow-ish
        "\033[38;5;191m",  # Yellow-ish Green
        "\033[38;5;114m",  # Olive Green
        "\033[38;5;26m",  # Dark Blue
    ]

    def __init(self):
        self.colors = []

    def get_colors(self, colormap, invert_colormap):
        """Returns a list of colors based on the colormap and invert_colormap
        arguments.
        """

        if colormap == "RdYlGn":
            self.colors = self.RdYlGn
        elif colormap == "BrBG":
            self.colors = self.BrBG
        elif colormap == "PiYG":
            self.colors = self.PiYG
        elif colormap == "PRGn":
            self.colors = self.PRGn
        elif colormap == "PiYG":
            self.colors = self.PiYG
        elif colormap == "PuOr":
            self.colors = self.PuOr
        elif colormap == "RdBu":
            self.colors = self.RdBu
        elif colormap == "RdGy":
            self.colors = self.RdGy
        elif colormap == "RdYlBu":
            self.colors = self.RdYlBu
        elif colormap == "Spectral":
            self.colors = self.Spectral
        else:
            raise ValueError(
                self.colormap
                + " is an incorrect colormap. Select one BrBG, PiYg, PRGn,"
                + " PuOr, RdBu, RdGy, RdYlBu, RdYlGn, or Spectral."
            )

        # check if we need to invert the colormap
        if invert_colormap:
            self.colors.reverse()

        return self.colors



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import cProfile
import traceback
import sys
import os

from datetime import datetime

try:
    from StringIO import StringIO  # python2
except ImportError:
    from io import StringIO  # python3
import pstats


def print_incomptable_msg(stats_file):
    """
    Function which makes the syntax cleaner in Profiler.write_to_file().
    """
    errmsg = """ Incompatible pstats file: {}\n Please run your code in Python {} to read in this file. """
    if sys.version_info[0] == 2:
        print(errmsg.format(stats_file, 3))
    if sys.version_info[0] == 3:
        print(errmsg.format(stats_file, 2.7))
    traceback.print_exc()


# class for profiling
class Profiler:
    """
    Wrapper class around cProfile.
    Exports a pstats file to be read by the pstats reader.
    """

    def __init__(self):
        self._prf = cProfile.Profile()
        self._output = "hatchet-profile"
        self._active = False

    def start(self):
        """
        Description: Place before the block of code to be profiled.
        """
        if self._active:
            print(
                "Start dissallowed in scope where profiler is running. Please add Profiler.stop() before attempting start."
            )
            raise

        self._active = True
        self._prf.enable()

    def stop(self):
        """
        Description: Place at the end of the block of code being profiled.
        """

        self._active = False
        self._prf.disable()
        self.write_to_file()

    def reset(self):
        """
        Description: Resets the profilier.
        """
        if self._active:
            print(
                "Reset dissallowed in scope where profiler is running. Please add Profiler.stop() before attempting reset."
            )
            raise

        self._prf = cProfile.Profile()

    def __str__(self):
        """
        Description: Writes stats object out as a string.
        """
        s = StringIO()
        pstats.Stats(self._prf, stream=s).print_stats()
        return s.getvalue()

    def write_to_file(self, filename="", add_pstats_files=[]):
        """
        Description: Write the pstats object to a binary
        file to be read in by an appropriate source.
        """
        sts = pstats.Stats(self._prf)

        if len(add_pstats_files) > 0:
            for stats_file in add_pstats_files:
                try:
                    sts.add(stats_file)
                except ValueError:
                    print_incomptable_msg(stats_file)
                    raise

        if filename == "":
            if os.path.exists(self._output + ".pstats"):
                now = datetime.now().strftime("%H%M%S")
                self.write_to_file(
                    "{}_{}.pstats".format(self._output, now), add_pstats_files
                )
            else:
                sts.dump_stats(self._output + ".pstats")
        else:
            if os.path.exists(filename):
                now = datetime.now().strftime("%H%M%S")
                filename = "{}_{}.pstats".format(filename, now)
            sts.dump_stats(filename)



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph

from abc import abstractmethod

# TODO The ABC class was introduced in Python 3.4.
# When support for earlier versions is (eventually) dropped,
# this entire "try-except" block can be reduced to:
# from abc import ABC
try:
    from abc import ABC
except ImportError:
    from abc import ABCMeta

    ABC = ABCMeta("ABC", (object,), {"__slots__": ()})


def _get_node_from_df_iloc(df, ind):
    node = None
    if isinstance(df.iloc[ind].name, tuple):
        node = df.iloc[ind].name[0]
    elif isinstance(df.iloc[ind].name, Node):
        node = df.iloc[ind].name
    else:
        raise InvalidDataFrameIndex(
            "DataFrame index elements must be either a tuple or a Node"
        )
    return node


def _get_parents_and_children(df):
    rel_dict = {}
    for i in range(len(df)):
        node = _get_node_from_df_iloc(df, i)
        if node not in rel_dict:
            rel_dict[node] = {}
            rel_dict[node]["parents"] = df.iloc[i].loc["parents"]
            rel_dict[node]["children"] = df.iloc[i].loc["children"]
    return rel_dict


def _reconstruct_graph(df, rel_dict):
    node_list = sorted(list(df.index.to_frame()["node"]))
    for i in range(len(df)):
        node = _get_node_from_df_iloc(df, i)
        if len(node.children) == 0:
            node.children = [node_list[nid] for nid in rel_dict[node]["children"]]
        if len(node.parents) == 0:
            node.parents = [node_list[nid] for nid in rel_dict[node]["parents"]]
    roots = [node for node in node_list if len(node.parents) == 0]
    return Graph(roots)


class DataframeReader(ABC):
    """Abstract Base Class for reading in checkpointing files."""

    def __init__(self, filename):
        self.filename = filename

    @abstractmethod
    def _read_dataframe_from_file(self, **kwargs):
        pass

    def read(self, **kwargs):
        df = self._read_dataframe_from_file(**kwargs)
        rel_dict = _get_parents_and_children(df)
        graph = _reconstruct_graph(df, rel_dict)
        graph.enumerate_traverse()
        exc_metrics = df.iloc[0, df.columns.get_loc("exc_metrics")]
        inc_metrics = df.iloc[0, df.columns.get_loc("inc_metrics")]
        default_metric = df.iloc[0, df.columns.get_loc("default_metric")]
        df.drop(
            columns=[
                "children",
                "parents",
                "exc_metrics",
                "inc_metrics",
                "default_metric",
            ],
            inplace=True,
        )
        return hatchet.graphframe.GraphFrame(
            graph,
            df,
            exc_metrics=exc_metrics,
            inc_metrics=inc_metrics,
            default_metric=default_metric,
        )


class InvalidDataFrameIndex(Exception):
    """Raised when the DataFrame index is of an invalid type."""



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from setuptools import setup, Extension
from Cython.Build import build_ext as cy_build_ext
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Get the version in a safe way which does not refrence hatchet `__init__` file
# per python docs: https://packaging.python.org/guides/single-sourcing-package-version/
version = {}
with open("./hatchet/version.py") as fp:
    exec(fp.read(), version)


setup(
    name="hatchet",
    version=version["__version__"],
    description="A Python library for analyzing hierarchical performance data",
    url="https://github.com/hatchet/hatchet",
    author="Abhinav Bhatele",
    author_email="bhatele@cs.umd.edu",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="",
    packages=[
        "hatchet",
        "hatchet.readers",
        "hatchet.util",
        "hatchet.external",
        "hatchet.tests",
        "hatchet.cython_modules.libs",
    ],
    setup_requires=["cython", "setuptools"],
    install_requires=[
        "pydot",
        "pyyaml",
        "matplotlib",
        "numpy",
        "pandas",
        "textx",
        "multiprocess",
        "caliper-reader",
        "pycubexr; python_version >= '3.6'",
    ],
    # TODO: the setup could be cleaner if we didn't dump the generated
    # .so files into _libs
    ext_modules=[
        Extension(
            "hatchet.cython_modules.libs.reader_modules",
            ["hatchet/cython_modules/reader_modules.pyx"],
        ),
        Extension(
            "hatchet.cython_modules.libs.graphframe_modules",
            ["hatchet/cython_modules/graphframe_modules.pyx"],
        ),
    ],
    cmdclass={"build_ext": cy_build_ext},
)
